# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import numpy as np
import numpy.ma as ma
import astropy.io.fits as fits
import copy
import synphot as syn
import astropy.units as u

from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS
from .custom_exceptions import EngineOutputError, EngineInputError
from .pandeia_warnings import report_warning_messages as warning_messages
from six.moves import range
from six.moves import zip
from . import debug_utils
from .utils import recursive_subclasses
from .projection import ProjectionFactory


def ReportFactory(calc_input, signal_list, noise_list, saturation_list, extracted, warnings, sub_str=''):

    """
    Function to take strategy and build/return a configured instance of the appropriate report. For ease of
    identifying and making targeted changes to a single strategy when necessary, we define subclasses for
    every strategy.

    Parameters
    ----------
    calc_input: dict
        The input calculation dictionary
    signal_list: list
        List of all the signal objects (one per dither)
    noise_list: list
        List of all the noise objects (one per dither)
    saturation_list: list
        List of all the saturation objects (one per dither)
    extracted: dict
        Dictionary of all the extracted observation products
    warnings: dict
        Dictionary of warning messages produced by the code
    sub_str: string
        String for tagging warnings from different subreports

    Returns
    -------
    cls: Report class object
        Properly configured Report class for the given strategy
    """

    method = calc_input['strategy']['method']

    # we need to run the API checks here after merging in the report-specific defaults
    #if webapp:
    #    self._api_checks(all_config)

    types = recursive_subclasses(Report)
    methods = [t.__name__.lower().replace('report', '') for t in types]
    type_map = dict(list(zip(methods, types)))

    if method not in methods:
        msg = "Unsupported or not yet implemented strategy: %s" % method
        raise EngineInputError(value=msg)
    else:
        cls = type_map[method](calc_input, signal_list, noise_list, saturation_list, extracted, warnings, sub_str)
        return cls


class Report(object):

    """
    This class contains functionality for generating reports and data products from the results
    of a Pandeia calculation.

    Parameters
    ----------
    calc_input: dict
        The engine API format dict used to configure the calculation
    signal_list : list of etc3D.DetectorSignal instances
        The calculated signal on the detector
    noise_list : list of etc3D.DetectorNoise instances
        The calculated noise on the detector
    extracted : dict
        The extracted data product generated by the Strategy
    warnings: dict
        Dict of warnings produced and collected through the course of the calculation
    sub_str: str
        String for tagging warnings from different subreports
    """

    def __init__(self, calc_input, signal_list, noise_list, saturation_list, extracted_list, warnings, sub_str):
        # For single pointing calculations, the signal and noise lists have one element.
        # However, handling calculations that involve multiple pointings is more tricky.
        # We want to grab one of the on_target calculations to get the relevant target
        # information for the report.
        if not isinstance(signal_list, list):
            raise EngineOutputError(value="Must provide Report signals and noises as lists.")

        self.r = dict()

        self.warnings = warnings
        # We use the calculation inputs for a few things.
        self.input = calc_input

        self.r['sub_reports'] = []

        if len(signal_list) == 1:
            self.signal = signal_list[0]
            self.noise = noise_list[0]
            self.saturation = saturation_list[0]
            self.extracted = extracted_list[0]
        elif len(signal_list) > 1:
            # Fish out the list of which ones are targets.
            """
            on_target = calc_input['strategy']['on_target']  # True if target, False otherwise
            target = [i for i, v in enumerate(on_target) if v][0]  # get first target
            """
            # The above method does not work in general. Assume the first plane is a target and
            # refactor when implementing dithers in general.
            target = 0

            self.signal = signal_list[target]
            self.noise = noise_list[target]
            self.saturation = saturation_list[target]
            self.extracted = extracted_list[target]
            # the zeroth element in the extracted_list is actually the combined results (where they exist) followed by
            # the same planes that are in the signal, noise, and saturation lists. We need to remove it so they line up.
            # if the extracted results are not from a multi-pointing mode, they are identical to the zeroth elements of
            # the signal, noise, and saturation arrays.
            extracted_list.pop(0)  # we want the subreports to have the individual subreport extractions
            subreports = range(len(signal_list))
            for signal, noise, saturation, extracted, subreport in zip(signal_list, noise_list, saturation_list, extracted_list, subreports):
                s = ReportFactory(calc_input, [signal], [noise], [saturation], [extracted], warnings, '{0:d}'.format(subreport+1))
                self.r['sub_reports'].append(s.as_dict())

        self.sub_str = sub_str

        # Get the full Signal and Noise objects for inspection.
        self.signals = signal_list
        self.noises = noise_list
        self.saturations = saturation_list
        self.extracts = extracted_list

        self.wave_index = 0
        self.projection_type = self.signal.projection_type
        self.projection = ProjectionFactory(self.projection_type)
        self._projection()

        # make a deep copy of the input so that nothing else can modify it later on
        self.r['input'] = copy.deepcopy(self.input)
        # initializing the rest of these as empty dictionaries to be filled
        self.r['1d'] = {}
        self.r['2d'] = {}
        self.r['3d'] = {}
        self.r['scalar'] = {}
        self.r['information'] = {}
        self.r['transform'] = {}

        # now fill the above dictionary sections with data
        self._scalar()
        self._1d()
        self._debugarrays()
        self._2d()
        self._3d()
        self._information()
        self._transform()
        self._warnings()


    def as_dict(self):
        """
        Functions to return the internally-created dictionary.
        It is possible to access the report dictionary attribute directly, but for backwards compatibility we've
        kept this function.

        Returns
        -------
        self.r : dict
            output dictionary of results
        """
        return self.r

    def as_fits(self):
        """
        Returns a dictionary of fits headers.
        This cannot be a separate function because it requires several attributes of the class.

        Returns
        -------
        self.f : dict
            output dictionary of fits tables
        """

        sub_reports = []
        if len(self.r['sub_reports']) > 1:
            for self.report in range(len(self.r['sub_reports'])):
                self.s = self.r['sub_reports'][self.report]
                self.f = dict()
                self.f['3d'] = {}
                self.f['2d'] = {}
                self.f['1d'] = {}
                self._3dfits()
                self._2dfits()
                self._1dfits()
                sub_reports.append(self.f)

        self.report = 0
        self.s = self.r
        self.f = dict()
        self.f['3d'] = {}
        self.f['2d'] = {}
        self.f['1d'] = {}
        self._3dfits()
        self._2dfits()
        self._1dfits()
        self.f['sub_reports'] = sub_reports

        return self.f

    def _set_wave(self, wref):
        """
        Given a wavelength wref, find and set the wavelength index to extract a slice.

        Parameters
        ----------
        wref: float
            Wavelength desired for the slice (in microns)
        """
        if wref is not None:
            if wref >= self.wave_pix.min() and wref <= self.wave_pix.max():
                self.wave_index = (np.abs(self.wave_pix - wref)).argmin()
            else:
                self.warnings['bad_waveref'] = warning_messages['bad_waveref'].format(wref,self.wave_pix.min(),
                                                                                  self.wave_pix.max(),
                                                                                  float(self.wave_pix[self.wave_index]))

    def _3dfits(self):
        """
        Format 3D outputs into fits headers.
        """
        for k in ['flux', 'flux_plus_background']:
            self.f['3d'][k] = []
            for i in range(len(self.s['3d'][k])):
                # the cube data coming out of pandeia.engine is ordered wavelength,x,y which is
                # backwards from the numpy convention of ordering axes from slowest to fastest.
                # it is set up this way because we need to broadcast 1D vectors (e.g. throughput
                # vs. wavelength) onto the 3D cubes. for this to work, wavelength must be the 3rd
                # axis. so flip the dimensions of the cube and then flip the Y axis to get the expected
                # orientation in the FITS data as viewed in DS9 or the like.
                o = fits.PrimaryHDU(np.moveaxis(self.s['3d'][k][i],2,0)[:, ::-1, :])
                # it should not be different, but pull the wcs info from the correct signal
                tbhdu, header = self.signals[self.report].cube_wcs_info()
                o.header.update(header)
                self.f['3d'][k].append(fits.HDUList([o, tbhdu]))

    def _2dfits(self):
        """
        Format 2D outputs into fits headers.
        """
        # this is set by _projection, and should be the same for all subreports of a single calculation
        detector_header = self.pix_grid.wcs_info()
        for k in self.s['2d']:
            # flip the Y axis to get the FITS data to look the
            # same in DS9 as in matplotlib.
            o = fits.PrimaryHDU(self.s['2d'][k][::-1, :])
            o.header.update(detector_header)
            self.f['2d'][k] = o

    def _1dfits(self):
        """
        Format 1D outputs into fits headers.
        """

        # These two are special one-array lists
        for k in ['wave_calc', 'wave_pix']:
            tbhdu = fits.BinTableHDU.from_columns([
                fits.Column(name='WAVELENGTH',
                            unit='um',
                            format="1D",
                            array=self.s['1d'][k])
            ])
            tbhdu.name = k
            self.f['1d'][k] = fits.HDUList([tbhdu])

        # now go through the rest
        # note that we're removing them from a list of dictionary keys, not removing them from the actual dictionary.
        results1d = list(self.s['1d'].keys())
        results1d.remove('wave_calc')
        results1d.remove('wave_pix')
        for k in results1d:
            tbhdu = fits.BinTableHDU.from_columns([
                fits.Column(name='WAVELENGTH',
                            unit='um',
                            format="1D",
                            array=self.s['1d'][k][0]),
                fits.Column(name=k,
                            format="1D",
                            array=self.s['1d'][k][1])
            ])
            tbhdu.name = k
            self.f['1d'][k] = fits.HDUList([tbhdu])

    def _projection(self):
        """
        Run the appropriate projection class. This is often overridden by strategies to include additional code.
        """
        self.grid, self.wave_pix, self.pix_grid = self.projection._project(self.signal, self.extracted)

    def _transform(self):
        """
        Spatial and wavelength information taken from the coordinate transform products
        """

        # set up the coordinate transform information that describes the image axes
        # for the 2D and 3D data.  first, get the spatial information from self.Grid

        self.r['transform'] = self.pix_grid.as_dict()

        # the wavelength sampling of the input spectrum is different
        # then the sampling by the detector. the model cubes that are
        # used to make the calculation use the input sampling. the detector
        # plane outputs for spectroscopy and IFU use the detector sampling

        # ## input model wavelength sampling ## #
        self.r['transform'].update(self.signal.spectral_model_transform())

        # ## detector plane wavelength sampling ## #
        self.r['transform'].update(self.signal.spectral_detector_transform())

        # the spectral mapping can be shifted by strategies (e.g. slitless). this is
        # reflected in wave_pix so update from there.
        self.r['transform']['wave_det_min'] = self.wave_pix.min()
        self.r['transform']['wave_det_max'] = self.wave_pix.max()
        self.r['transform']['wave_det_refval'] = self.wave_pix[0]

    def _scalar(self):
        """
        Scalar data products

        Note that this method allows for certain variables to be unspecified in the inputs, on a case by case basis.
        This permissiveness is an intentional engine feature - parameters that are not given in the inputs are to be
        filled in by (sensible, astronomer-verified) defaults, and should not cause crashes. #3737
        """

        # get areas in pixels of source and background regions
        self.r['scalar']['background_area'] = self.extracted['background_area']
        self.r['scalar']['extraction_area'] = self.extracted['extraction_area']
        self.extraction_area = self.extracted['extraction_area']
        self.background_area = self.extracted['background_area']

        self.r['scalar']['fraction_saturation'] = self.extracted['saturation_products']['fraction_saturation']
        self.r['scalar']['sat_ngroups'] = np.min(self.extracted['detector_ngroups'])
        self.r['scalar']['brightest_pixel'] = self.signal.brightest_pixel

        # put some instrument and strategy configuration info into the scalar report
        self.r['scalar']['filter'] = self.input['configuration']['instrument']['filter']
        self.r['scalar']['disperser'] = self.input['configuration']['instrument']['disperser']

        # target_xy is not found in SOSS or MSAFullApPhot, and may not be found in future strategies as well.
        # If it's not there, assume the target position is the center of the FOV, i.e. (0.0, 0.0).
        if 'target_xy' in self.input['strategy']:
            self.r['scalar']['x_offset'] = self.input['strategy']['target_xy'][0]
            self.r['scalar']['y_offset'] = self.input['strategy']['target_xy'][1]
        else:
            self.r['scalar']['x_offset'] = 0.0
            self.r['scalar']['y_offset'] = 0.0

        # Not all strategies define an aperture size either. SOSS doesn't because the size is defined by an
        # the grism. TAPhot and TACentroid do not require it to be specified, as it is fixed and cannot be changed
        # by users. Optimal extraction strategies likely will not use this, either, once implemented.
        if 'aperture_size' in self.input['strategy']:
            self.r['scalar']['aperture_size'] = self.input['strategy']['aperture_size']
        else:
            self.r['scalar']['aperture_size'] = "N/A"

    def _3d(self):
        """
        Compute 3D data products
        """

        # this is the data cube of the input signal
        self.flux = self.signal.flux_cube_list
        self.flux_plus_bg = self.signal.flux_plus_bg_list
        self.r['3d']['flux'] = self.flux

        # this is the data cube of the input signal plus background
        self.r['3d']['flux_plus_background'] = self.flux_plus_bg

    def _2d(self):
        """
        Compute 2D data products. Though some strategies override this, these are mostly projection-dependent
        """
        resultdict,self.bg_pix = self.projection._2d(self.extracted, self.saturation, self.signal)
        self.r['2d'].update(resultdict)

    def _1d(self):
        """
        Compute 1D data products, and the scalar values that are extracted from the 1D products.

        NOTE: We do not do any NaN checking here.  This may cause certain values like SNR
        to be reported as NaN in 1d, but as 0 in scalars.
        """

        # this is the wavelength sampling of the calculation
        self.wave = self.signal.wave

        self.r['1d']['wave_pix'] = self.wave_pix
        self.r['1d']['wave_calc'] = self.wave

        # Target spectrum
        self.r['1d']['target'] = [self.wave, self.signal.total_flux]

        # Transmission/focal plane rate
        self.r['1d']['fp'] = [self.wave, self.signal.fp_rate.sum(axis=(0, 1))]

        # input background
        self.r['1d']['bg'] = [self.wave, self.signal.background.MJy_sr]

        # Background rate
        self.r['1d']['bg_rate'] = [self.wave, self.signal.bg_fp_rate]

        # Values calculated by the strategy
        sn = self.extracted['extracted_flux'] / self.extracted['extracted_noise']

        self.r['1d']['sn'] = [self.wave_pix, sn]
        self.r['1d']['extracted_noise'] = [self.wave_pix, self.extracted['extracted_noise']]
        self.r['1d']['extracted_flux'] = [self.wave_pix, self.extracted['extracted_flux']]
        self.r['1d']['extracted_flux_plus_bg'] = [self.wave_pix, self.extracted['extracted_flux_plus_bg']]

        self.r['1d']['total_flux'] = [self.wave_pix, self.extracted['source_flux_in_fov']]

        self.r['1d']['extracted_bg_total'] = [self.wave_pix, self.extracted['extracted_bg_total']]
        self.r['1d']['extracted_bg_only'] = [self.wave_pix, self.extracted['extracted_bg_only']]
        self.r['1d']['extracted_contamination'] = [self.wave_pix, self.extracted['extracted_bg_total'] - self.extracted['extracted_bg_only']]
        self.r['1d']['n_partial_saturated'] = [self.wave_pix, self.extracted['saturation_products']['partial']]
        self.r['1d']['n_full_saturated'] = [self.wave_pix, self.extracted['saturation_products']['full']]

        self.curves = self.r['1d']

        # if there's a detector gap, the noise is 0 and sn goes to np.inf. fix that here
        # so that we can get a proper maximum S/N and report scalar values there.
        # NOTE: this value is only used in the scalar values, the 1d sn
        # may still contain a NaN.
        sn = ma.fix_invalid(sn, fill_value=0.0).data

        rw = float(self.wave_pix[self.wave_index])
        self.r['scalar']['sn'] = float(sn[self.wave_index])
        self.r['scalar']['extracted_flux'] = float(self.extracted['extracted_flux'][self.wave_index])

        # if the SN is 0.0 (e.g. due to heavy saturation), we do not know what the variance is and set it to np.nan
        if np.abs(self.r['scalar']['sn']) > 0.0:
            self.r['scalar']['extracted_noise'] = float(self.extracted['extracted_noise'][self.wave_index])
        else:
            self.r['scalar']['extracted_noise'] = np.nan

        self.r['scalar']['background_total'] = float(self.extracted['extracted_bg_total'][self.wave_index])
        self.r['scalar']['background_sky'] = float(self.extracted['extracted_bg_only'][self.wave_index])

        # only check for contamination if background level is above some epsilon value. otherwise round-off
        # errors can cause unstable test results.
        if np.abs(self.r['scalar']['background_total']) > 1.0e-9:
            self.r['scalar']['contamination'] = (self.r['scalar']['background_total'] - self.r['scalar']['background_sky']) / \
                                                self.r['scalar']['background_total']
        else:
            self.r['scalar']['contamination'] = 0.0

        self.r['scalar']['reference_wavelength'] = rw
        self.r['scalar']['background'] = self.signal.background.bg_spec(rw * PANDEIA_WAVEUNITS, flux_unit=u.MJy).value

    def _information(self):
        """
        Collect some additional information that is not the product of a calculation.
        """

        # get the ExposureSpecification.
        self.exposure_specification = self.signal.current_instrument.get_exposure_pars()

        self.r['information']['calc_type'] = self.signal.projection_type

        self.r['information']['exposure_specification'] = self.exposure_specification.__dict__

        # Some strategies like IFU and coronagraphy add dithers. Some are on-source, others are off-source.
        # This ensures that all on-source dithers are included in the "total_exposure_time", whereas
        # off-source time is only reported in "all_dithers_time".
        total_exposure_time = self.r['information']['exposure_specification']['total_exposure_time']
        if 'n_on_source' in self.extracted:
            self.r['scalar']['total_exposure_time'] = total_exposure_time * self.extracted['n_on_source']
            self.r['scalar']['all_dithers_time'] = total_exposure_time * self.extracted['n_total']
        else:
            self.r['scalar']['total_exposure_time'] = total_exposure_time
            self.r['scalar']['all_dithers_time'] = total_exposure_time

        # Report on measurement, exposure time (time for a single exposure) and saturation time.
        # The r['scalar']['exposure_time'] value reported back to the client is not used at this point.
        self.r['scalar']['exposure_time'] = self.r['information']['exposure_specification']['exposure_time']
        self.r['scalar']['measurement_time'] = self.r['information']['exposure_specification']['measurement_time']
        self.r['scalar']['saturation_time'] = self.r['information']['exposure_specification']['saturation_time']
        self.r['scalar']['total_integrations'] = self.r['information']['exposure_specification']['total_integrations']

        # Report the exposure duty cycle
        self.r['scalar']['duty_cycle'] = self.r['information']['exposure_specification']['duty_cycle']

        # get the pixel CR rate in events/s and convert to events/ramp
        self.r['scalar']['cr_ramp_rate'] = self.noise.pix_cr_rate * self.r['information']['exposure_specification'][
            'saturation_time']

    def _warnings(self, sub_words={'1':'Dither','2':'Dither'}):
        """
        Collect warnings, and generate saturation warnings

        This module contains both the collected warnings from elsewhere in the program, and also the saturation
        warnings (which are, currently, the only warnings that may be different between subreports)

        Parameters
        -----------
        sub_words : dict
            dict of strings to use when differentiating warning subreports
        """

        # The only time the saturation warnings do NOT come from self.extracted['detector_saturation'] is with the
        # IFU modes, which override this method.
        sat_obj = self.extracted['detector_saturation']
        self.r['warnings'] = self.warnings

        self.n_partial_saturated = (sat_obj == 1).sum()
        self.n_full_saturated = (sat_obj == 2).sum()

        # Saturation warnings can be different in different subreports
        # This logic works because subreports are processed first - if they exist, the main (combined) report's warning
        # is not triggered.
        if self.n_partial_saturated > 0:
            if self.sub_str != '':
                self.r['warnings'][self.sub_str+'partial_saturated'] = '{0:}: {1:}'.format(sub_words[self.sub_str], warning_messages['partial_saturated'].format(self.n_partial_saturated))
            else:
                if '1partial_saturated' not in self.r['warnings']:
                    self.r['warnings']['partial_saturated'] = warning_messages['partial_saturated'].format(self.n_partial_saturated)
        if self.n_full_saturated > 0:
            if self.sub_str != '':
                self.r['warnings'][self.sub_str+'full_saturated'] = '{0:}: {1:}'.format(sub_words[self.sub_str], warning_messages['full_saturated'].format(self.n_full_saturated))
            else:
                if '1full_saturated' not in self.r['warnings']:
                    self.r['warnings']['full_saturated'] = warning_messages['full_saturated'].format(self.n_full_saturated)

    def _debugarrays(self):
        """
        Collect the debug information if it exists
        """
        if debug_utils.debugarrays and debug_utils.debugarrays.get_report():
            self.r['debugarrays'] = debug_utils.debugarrays.get_report()


class CoronagraphyReport(Report):
    """
    This report class is specifically for Coronagraphy (and its derivatives)
    Parameters are read in through Report.__init__

    """

    def _1d(self):
        """
        Coronagraphy-specific contrast quantities
        """
        Report._1d(self)
        # Coronagraphy always has contrast curves
        self.r['1d']['contrast'] = self.extracted['contrast_curve']
        self.curves['contrast'] = self.extracted['contrast_curve']

    def _1dfits(self):
        Report._1dfits(self)
        # contrast is NOT like the other 1D products and must be handled separately (overwritten).
        tbhdu = fits.BinTableHDU.from_columns([
            fits.Column(name='SEPARATION',
                        unit='arcsec',
                        format="1D",
                        array=self.s['1d']['contrast'][0]),
            fits.Column(name="CONTRAST",
                        format="1D",
                        array=self.s['1d']['contrast'][1])
        ])
        tbhdu.name = 'contrast'
        self.f['1d']['contrast'] = fits.HDUList([tbhdu])

    def _2d(self):
        """
        Coronagraphy-specific 2D calculations.

        This does not call any of the usual _2d() infrastructure (in projection.py classes) because Coronagraphy
        2D planes are different from any other mode. The coronagraphy strategy produces (and uniquely, uses) a
        detector_signal that does not have background added in. This is presumably equivalent to an astronomer
        subtracting the background off of their images before scaling and subtracting the PSF subtraction source.
        """

        # Signal and noise in 2D. Get from extracted products
        s = self.extracted['detector_signal']
        n = self.extracted['detector_noise']
        self.detector_saturation = self.extracted['detector_saturation']

        # not reported, but needs to be available as a class attribute
        self.bg_pix = self.signal.rate_plus_bg - self.signal.rate

        # The noise is undefined if the pixel is fully saturated.
        # The engine currently returns noise=0 if the pixel has full saturation, which is
        # confusing since the noise is not 0, but rather undetermined or infinite. Setting
        # the noise to NaN ensures that the S/N of saturated pixels are NaNs.
        n[self.detector_saturation == 2] = np.nan

        self.detector_sn = s / n
        self.detector_signal = s + n * np.random.randn(n.shape[0], n.shape[1])

        self.r['2d']['detector'] = self.detector_signal
        self.r['2d']['snr'] = self.detector_sn
        self.r['2d']['saturation'] = self.detector_saturation
        self.r['2d']['ngroups_map'] = self.extracted['detector_ngroups']

    def _scalar(self):
        """
        Coronagraphy-specific scalar quantities.

        By design, engine inputs should be optional. That should include this contrast information, too.
        """

        Report._scalar(self)
        if 'contrast_separation' in self.input['strategy']:
            contrast_separation = self.input['strategy']['contrast_separation']
            self.r['scalar']['contrast_separation'] = contrast_separation
            self.r['scalar']['contrast_azimuth'] = self.input['strategy']['contrast_azimuth']
            self.r['scalar']['contrast'] = np.interp(contrast_separation, self.extracted['contrast_curve'][0],
                                                self.extracted['contrast_curve'][1])

    def _warnings(self):
        Report._warnings(self, sub_words={'1':'Science Scene', '2':'PSF Subtraction Source', '3':'Contrast Observation'})

class SpecApPhotReport(Report):
    """
    This report class is specifically for Aperture Spectral Extraction
    Parameters are read in through Report.__init__

    """

    def _projection(self):
        """
        Aperture Spectral Extraction-specific projection.

        SpecApPhot requires a reference wavelength (in the middle of the spectral coverage range, if the input is None)
        to extract scalar properties from 1D products.
        """
        Report._projection(self)

        self.wave_index = int(len(self.wave_pix) / 2.0)
        if 'reference_wavelength' in self.input['strategy']:
            wref = self.input['strategy']['reference_wavelength']
            self._set_wave(wref)

class MSAFullApPhotReport(SpecApPhotReport):
    """
    This report class is specifically for MSA Full Aperture Spectroscopy
    Parameters are read in through Report.__init__

    """
    pass


class MSAShutterApPhotReport(Report):
    """
    This report class is specifically for MSA Shutter Aperture Photometry
    Parameters are read in through Report.__init__

    """
    pass


class MSAApPhotReport(Report):
    """
    This report class is specifically for MSA Shutter Aperture Photometry
    Parameters are read in through Report.__init__

    """
    pass


class SOSSReport(SpecApPhotReport):
    """
    This report class is specifically for SOSS
    Parameters are read in through Report.__init__

    """
    pass


class IFUApPhotReport(Report):
    """
    This report class is specifically for IFU Aperture Photometry (and its derivatives)
    Parameters are read in through Report.__init__

    """

    def _information(self):
        """
        IFU Aperture Photometry-specific information.

        IFU modes are computed as spectroscopy, but reported as imaging (reconstructed data cubes)
        """
        Report._information(self)
        # this is used by the UI to decide how to make the plots. even though IFUs are spectroscopic,
        # they're plotted as 2D spatial images. they could be shown as either single planes from the
        # reconstructed cubes or some combination of multiple planes.
        self.r['information']['calc_type'] = "image"

    def _projection(self):
        """
        IFU Aperture Photometry-specific projection.

        IFU modes use extracted properties for their projection, and also require wavelength setting in a fashion
        similar to Aperture Spectral Extraction.
        """
        self.grid = self.signal.grid
        # for IFUs we want the reconstructed image plane that's build by the strategy
        self.pix_grid = self.extracted['plane_grid']
        self.wave_pix = self.extracted['wavelength']

        self.wave_index = int(len(self.wave_pix) / 2.0)
        if 'reference_wavelength' in self.input['strategy']:
            wref = self.input['strategy']['reference_wavelength']
            self._set_wave(wref)

    def _2d(self):
        """
        IFU Aperture Photometry-specific 2D products.

        For IFU modes, all of the 2d outputs of IFUs are linked to the 3D cubes and defined there.
        """
        pass


    def _3d(self):
        """
        IFU Aperture Photometry-specific 3D and 2D products.

        IFU modes report 3D data cubes, in addition to the input model cubes all other strategies report.
        The 2D IFU data is taken from slices of these 3D cubes.
        """
        Report._3d(self)

        # adding a np.random.randn() instance here operating on self.extracted['detector_signal'], before it's run for
        # cube_sim, will get the same random numbers as the v1.2 and earlier version of Report.py
        # Signal and noise. Get from extracted products

        self.cube_signal, self.cube_noise, self.cube_saturation, self.cube_group_map, self.cube_plane_grid = self.extracted['reconstructed']
        self.cube_sim = self.cube_signal + self.cube_noise * np.random.randn(*self.cube_noise.shape)

        self.r['3d']['reconstructed'] = self.cube_sim
        self.r['3d']['reconstructed_signal'] = self.cube_signal
        self.r['3d']['reconstructed_noise'] = self.cube_noise
        self.r['3d']['reconstructed_saturation'] = self.cube_saturation
        self.r['3d']['reconstructed_ngroups_map'] = self.cube_group_map
        self.r['3d']['reconstructed_snr'] = self.cube_signal / self.cube_noise

        # use cube planes for IFUs, though this may be temporary. the actual detector image for an IFU
        # observation is a set of spectra, one for each IFU slice. Populate the detector and SNR images from the
        # planes of the reconstructed cubes, but collapse the saturation cube to show the most severe saturation
        # for each spatial pixel.

        self.r['2d']['detector'] = self.cube_sim[self.wave_index, :, :]
        self.r['2d']['snr'] = self.cube_signal[self.wave_index, :, :] / self.cube_noise[self.wave_index, :, :]
        self.r['2d']['saturation'] = np.amax(self.cube_saturation, axis=0)
        self.r['2d']['ngroups_map'] = np.amin(self.cube_group_map, axis=0)
        # This is the background rate in each pixel without sources
        self.bg_pix = self.signal.rate_plus_bg - self.signal.rate

    def _3dfits(self):
        """
        IFU Aperture Photometry-specific 3D fits output.

        IFU modes report 3D data cubes, in addition to the input model cubes all other strategies report.
        """
        Report._3dfits(self)
        header = self.pix_grid.wcs_info()
        t = self.signal.spectral_detector_transform()
        header['ctype3'] = 'Wavelength'
        header['crpix3'] = 1
        header['crval3'] = t['wave_det_min'] - 0.5 * t['wave_det_step']
        header['cdelt3'] = t['wave_det_step']
        header['cunit3'] = 'um'
        header['cname3'] = 'Wavelength'
        for k in ['', '_signal', '_noise', '_snr', '_saturation']:
            key = 'reconstructed' + k
            # the reconstructed cubes are in the proper z, y, x order...
            # The Y axis must be flipped so the result looks the same in DS9 (etc) as
            #  it does in the client.
            o = fits.PrimaryHDU(self.r['3d'][key][:, ::-1, :])
            o.header.update(header)
            self.f['3d'][key] = o

    def _warnings(self):
        """
        Collect warnings for the IFU mode, and generate saturation warnings. This does not call to Report._warnings()
        because ALL of the saturation warnings for IFU modes are different from anywhere else.

        Parameters
        -----------
        sub_words : dict
            dict of strings to use when differentiating warning subreports
        """
        sub_words = {'1': 'Nod 1', '2': 'Nod 2'}
        self.r['warnings'] = self.warnings
        sat_obj = self.cube_saturation[self.wave_index]

        self.n_partial_saturated = (sat_obj == 1).sum()
        self.n_full_saturated = (sat_obj == 2).sum()

        # Saturation warnings can be different in different subreports
        # This logic works because subreports are processed first - if they exist, the main (combined) report's warning
        # is not triggered.

        # These are the warnings about the specific requested slice of the cube. Text is different from the usual
        # partial_saturation or full_saturation warning.
        if self.n_partial_saturated > 0:
            if self.sub_str != '':
                self.r['warnings'][self.sub_str+'ifu_partial_saturated'] = '{0:}: {1:}'.format(sub_words[self.sub_str], warning_messages['ifu_partial_saturated'].format(self.n_partial_saturated))
            else:
                if '1ifu_partial_saturated' not in self.r['warnings']:
                    self.r['warnings']['ifu_partial_saturated'] = warning_messages['ifu_partial_saturated'].format(self.n_partial_saturated)
        if self.n_full_saturated > 0:
            if self.sub_str != '':
                self.r['warnings'][self.sub_str+'ifu_full_saturated'] = '{0:}: {1:}'.format(sub_words[self.sub_str], warning_messages['ifu_full_saturated'].format(self.n_full_saturated))
            else:
                if '1ifu_full_saturated' not in self.r['warnings']:
                    self.r['warnings']['ifu_full_saturated'] = warning_messages['ifu_full_saturated'].format(self.n_full_saturated)

        n_cube_partial = (self.cube_saturation == 1).sum()
        n_cube_full = (self.cube_saturation == 2).sum()

        # these are the warnings about the entire cube
        if n_cube_partial > 0:
            if self.sub_str != '':
                self.r['warnings'][self.sub_str+'cube_partial'] = '{0:}: {1:}'.format(sub_words[self.sub_str], warning_messages['cube_partial'].format(n_cube_partial))
            else:
                if '1cube_partial' not in self.r['warnings']:
                    self.r['warnings']['cube_partial'] = warning_messages['cube_partial'].format(n_cube_partial)
        if n_cube_full > 0:
            if self.sub_str != '':
                self.r['warnings'][self.sub_str+'cube_full'] = '{0:}: {1:}'.format(sub_words[self.sub_str], warning_messages['cube_full'].format(n_cube_full))
            else:
                if '1cube_full' not in self.r['warnings']:
                    self.r['warnings']['cube_full'] = warning_messages['cube_full'].format(n_cube_full)

class IFUNodInSceneReport(IFUApPhotReport):
    """
    This report class is specifically for IFUNodInScene (and its derivatives)
    Parameters are read in through Report.__init__

    """
    pass

class IFUNodOffSceneReport(IFUNodInSceneReport):
    """
    This report class is specifically for IFUNodOffScene (and its derivatives)
    Parameters are read in through Report.__init__

    """
    pass


class ImagingApPhot(Report):
    """
    This report class is specifically for Imaging Aperture Photometry (and its derivatives)
    Parameters are read in through Report.__init__

    """
    def _warnings(self, sub_words={'1':'Dither','2':'Dither'}):
        """
        Collect warnings, and generate saturation warnings

        This module contains both the collected warnings from elsewhere in the program, and also the saturation
        warnings (which are, currently, the only warnings that may be different between subreports)

        Parameters
        -----------
        sub_words : dict
            dict of strings to use when differentiating warning subreports
        """

        Report._warnings(self)
        if (self.input['configuration']['instrument']['instrument'] == "niriss") and (self.input['configuration']['instrument']['mode'] == "ami"):

            # Saturation warnings can be different in different subreports
            # This logic works because subreports are processed first - if they exist, the main (combined) report's warning
            # is not triggered.
            if self.n_partial_saturated > 0:
                if '1partial_saturated' not in self.r['warnings']:
                        self.r['warnings']['partial_saturated'] = warning_messages['partial_saturated_ami'].format(self.n_partial_saturated)
            if self.n_full_saturated > 0:
                if '1full_saturated' not in self.r['warnings']:
                        self.r['warnings']['full_saturated'] = warning_messages['full_saturated_ami'].format(self.n_full_saturated)


class TAPhotReport(Report):
    """
    This report class is specifically for Target Acquisition Photocenter (and its derivatives)
    Parameters are read in through Report.__init__

    """
    def _warnings(self):
        """
        TAPhot-specific warnings.

        Target Acquisition modes are uniquely concerned with saturation, because target acquisition is not possible
        if too many pixels on the detector are saturated. Likewise, the SNR must be higher than a certain value for
        target acquisition to succeed.
        """
        Report._warnings(self)
        # Target acquisitions allow some number of saturated pixels within the extraction aperture. We
        # count both the partial and full saturated pixels.
        number_saturated_pixels = self.extracted['saturation_products']['full'][0] + self.extracted['saturation_products']['partial'][0]
        if number_saturated_pixels > self.signal.current_instrument.max_saturated_pixels:
            self.warnings['ta_max_saturated_pixels'] = warning_messages['ta_max_saturated_pixels'].format(number_saturated_pixels,
                    self.signal.current_instrument.max_saturated_pixels)

        # Compare to the target acquisition SNR requirement. If target is fully saturated, there is no need to
        # check SNR (since it will be 0)
        sn = self.extracted['extracted_flux'] / self.extracted['extracted_noise']
        # if there's a detector gap, the noise is 0 and sn goes to np.inf. fix that here
        # so that we can get a proper maximum S/N and report scalar values there.
        # NOTE: this value is only used in the scalar values, the 1d sn
        # may still contain a NaN.
        sn = ma.fix_invalid(sn, fill_value=0.0).data
        if (sn[self.wave_index] < self.signal.current_instrument.min_snr_threshold) & \
                (not self.extracted['saturation_products']['full'][0] > 0):
            self.warnings['ta_snr_threshold'] = warning_messages['ta_snr_threshold'].format(sn[self.wave_index],
                                                                     self.signal.current_instrument.min_snr_threshold)


class TACentroidReport(TAPhotReport):
    """
    This report class is specifically for Target Acquisition Centroid
    Parameters are read in through Report.__init__

    """
    def _1d(self):
        """
        TACentroid-specific 1D products.

        Some strategies may not return a flux. TargetAcqCentroid is one example for now; it reports the centroid in place
        of the flux. The property is still called "extracted_flux" in the curves, but we rename them here to avoid
        confusion. The "extracted_noise" is still the standard deviation of the extracted product.
        """
        Report._1d(self)

        self.r['scalar']['extracted_centroid'] = float(self.extracted['extracted_flux'][self.wave_index])
        self.r['scalar']['extracted_flux'] = np.nan



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

from .custom_exceptions import EngineInputError
from .utils import recursive_subclasses, merge_data
from .instrument import Instrument

# need to import the Instrument subclasses from whereever they're defined
from .jwst import NIRSpec, NIRCam, NIRISS, MIRI
from .roman import WFI
from .hst import WFC3IR, WFC3UVIS, ACS, STIS, COSFUV

def InstrumentFactory(config={}, webapp=False, **kwargs):
    """
    Function to take configuration data and build/return an appropriately configured
    instance of the desired Instrument subclass.

    Parameters
    ----------
    config: dict
        Configuration data in engine API format.  Used to suss out which instrument to configure
        and then passed along.
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    all_config = merge_data(config, dict(**kwargs))
    types = recursive_subclasses(Instrument)
    instruments = [t.__name__.lower() for t in types]
    inst_map = dict(list(zip(instruments, types)))

    # get the instrument name out of the input configuration
    try:
        instrument = all_config['instrument']['instrument']
    except KeyError as e:
        msg = "Must provide Instrument name via engine API compatible dict or instrument=<name> keyword/value pair."
        raise EngineInputError(value=msg)

    # get the instrument mode out of the input configuration
    try:
        mode = all_config['instrument']['mode']
    except KeyError as e:
        msg = "Must provide Instrument mode via engine API compatible dict or mode=<modename> keyword/value pair. (%s)" % e
        raise EngineInputError(value=msg)

    # make sure requested instrument is one we actually support
    if instrument not in instruments:
        msg = "Instrument %s not supported/implemented." % instrument
        raise EngineInputError(value=msg)
    else:
        cls = inst_map[instrument](mode=mode, config=config, webapp=webapp, **kwargs)
        return cls



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# This is the new Roman Jupyter Notebook GUI - it should replace the previous notebook utils, which
# were written for a much older version of iPyWidgets.

from __future__ import division, absolute_import

import os
import json

import numpy as np
import matplotlib
matplotlib.use('nbagg')
import matplotlib.pyplot as plt

from pandeia.engine.perform_calculation import perform_calculation
from pandeia.engine.io_utils import read_json, write_json, NumPyArangeEncoder
from pandeia.engine.calc_utils import build_default_calc, build_default_source

from ipywidgets import widgets, GridspecLayout
from IPython.display import display, clear_output
import traitlets

refdata = os.environ['pandeia_refdata']

MAX_SOURCES=20


def get_config(filename):
    """
    read configuration data from a JSON file and create a dict using display_strings as the keys.
    these dicts will be used to populate pull-downs with descriptive names, but provide lookup to
    the values that need to be passed back to the engine.

    Parameters
    ----------
    filename: string
        filename of JSON config file to Load

    Returns
    -------
    conf: dict
        JSON configuration data with keys swapped out for display_strings where available.
    """
    conf_data = read_json(filename)
    conf = {}
    for k in conf_data:
        if "display_string" in conf_data[k]:
            conf[conf_data[k]["display_string"]] = k
        else:
            conf[k] = k
    return conf

dummy = None
geom_mapping = {'Point Source': 'point', 'Flat':'flat', 'Power Law': 'power', '2D Gaussian': 'gaussian2d',
                'Sersic (Effective Radius)': 'sersic', 'Sersic (Scale Radius)': 'sersic_scale'}
norm_mapping = {'infinity': 'integ_infinity', 'scale radius': 'surf_scale', 'center': 'surf_center'}

phoenix_config_file = os.path.join(refdata, 'sed', 'phoenix', 'spectra.json')
PHOENIX_CONFIG = get_config(phoenix_config_file)
gal_config_file = os.path.join(refdata, 'sed', 'brown', 'spectra.json')
GAL_CONFIG = get_config(gal_config_file)
star_config_file = os.path.join(refdata, 'sed', 'hst_calspec', 'spectra.json')
STAR_CONFIG = get_config(star_config_file)
extinction_config_file = os.path.join(refdata, 'extinction', 'config.json')
with open(extinction_config_file) as extfile:
    extinction_config = json.load(extfile)
EXTINCTION_LAWS = [extinction_config['laws'][key]['display_string'] for key in extinction_config['laws']]
EXTINCTION_BANDS = [key for key in extinction_config['bandpasses']]

extinction_mapping = {extinction_config['laws'][key]['display_string']: key for key in extinction_config['laws']}

"""
Functions to aid in creating valid tests and identifying valid instrument configurations.
"""


class Instrument():
    """
    Class for parsing valid instrument configurations from config.json files.
    This is primarily in aid of generating tests, but can be used for anything
    that needs to find valid instrument configurations.

    Given a telescope and instrument, class objects will expose the available
     modes.
    If a mode is set, the class object will expose available apertures and
     strategies for that mode.
    If an aperture is set as well, the class object will expose the valid
     filters, dispersers, readout patterns, and subarrays,

    The remainder of the instrument configuration is available as self.config

    Parameters
    ----------
    instrument; string
        One of the instruments defined in pandeia_refdata.
    telescope: string
        Telescope defined in pandeia_refdata. Currently 'jwst', 'roman',
        or 'hst'
    webapp: bool
        Change this attribute to True to enable stricter filter lists
    """

    def __init__(self,instrument, telescope):
        self.instrument = instrument
        self.telescope = telescope
        self.data_dir = os.environ['pandeia_refdata']
        config_file = open('{0:}/{1:}/{2:}/config.json'.format(self.data_dir,self.telescope,self.instrument))
        self.config = json.load(config_file)
        self.webapp = False
        self.get_modes()

    def get_modes(self):
        self.modes = self.config['modes']

    def set_mode(self, mode):
        self.mode = mode
        self.get_apertures()
        self.get_strategies()

    def set_element(self, element, value):
        setattr(self,element,value)

    def get_apertures(self):
        self.apertures_all = self.config['apertures']

        if 'apertures' in self.config['mode_config'][self.mode]:
            self.apertures = self.config['mode_config'][self.mode]['apertures']
        else:
            self.apertures = self.apertures_all

    def set_aperture(self,aperture):
        self.aperture = aperture
        self._lookup('dispersers')
        self._lookup('filters')
        self._lookup('subarrays')
        self._lookup('readout_patterns')

    def _lookup(self,element):
        setattr(self,'{0:}_all'.format(element),self.config[element])

        if element in self.config['mode_config'][self.mode]:
            setattr(self, element, self.config['mode_config'][self.mode][element])
        else:
            # if it's not found in mode_config, use the base list
            setattr(self, element, getattr(self,'{0:}_all'.format(element)))

        # Web constraints will remove some more items
        if self.webapp:
            self.constrain(element=element)

        if getattr(self,element) == []:
            setattr(self, element, [None])

    def get_strategies(self):
        self.strategies = self.config['strategy_config'][self.mode]['permitted_methods']

    def constrain(self, element=None, constrain_on='aperture'):
        # Config Constraints lists dispersers in one of four ways:
        # Under apertures/dispersers/<mode>
        # Under apertures/dispersers/default if it applies to all modes but a particular one
        # Under apertures/dispersers if it applies to ALL modes
        # If unlisted, use the mode or global values.

        # NIRSpec also constrains by disperser.

        survivors = None
        constraint = '{0:}s'.format(constrain_on)
        # Three things required to do this:
        # 1. The configuration dictionary must have a config_constraints section
        # 2. The element to constrain on must be defined (for example, aperture)
        # 3. The element to constrain on must be in the config_constraints section (as, example, apertures)
        if 'config_constraints' in self.config:
            if hasattr(self,constrain_on):
                const = getattr(self,constrain_on)
                if constraint in self.config['config_constraints']:
                    if const in self.config['config_constraints'][constraint]:
                        if element in self.config['config_constraints'][constraint][const]:
                            if self.mode in self.config['config_constraints'][constraint][const][element]:
                                survivors = self.config['config_constraints'][constraint][const][element][self.mode]
                            elif "default" in self.config['config_constraints'][constraint][const][element]:
                                survivors = self.config['config_constraints'][constraint][const][element]['default']
                            else:
                                survivors = self.config['config_constraints'][constraint][const][element]

        # only actually set a new list if there's something worth setting.
        if survivors is not None:
            setattr(self,element,survivors)

class SourceObj(object):
    """
    A source definition block. It's comprised of three main blocks:
    Flux
    SED
    Advanced

    The Advanced block has three sub-blocks:
    Geometry
    Position
    Redshift and Extinction
    """
    def __init__(self):
        self.container = widgets.VBox(width="100%", background_color="#EEEEEE")

        # default is to hide source panels
        self.active = False
        self.container.layout.display = 'none'
        # Flux
        self.flux_box = widgets.HBox(padding='10px', width="100%")
        self.flux = widgets.BoundedFloatText(description="Source flux: ", min=0.0, max=1.0e30, value=23.0, step=0.1)
        self.funits = widgets.Dropdown(value='abmag', options=['abmag', 'njy', 'ujy', 'mjy', 'jy'])
        atwave = widgets.HTML(value=" at ", margin='5px')
        self.wave = widgets.BoundedFloatText(min=0.1, max=99999.0, step=0.1, value=1.5)
        waveunits = widgets.HTML(value="microns", margin='5px')
        self.flux_box.children = [self.flux, self.funits, atwave, self.wave, waveunits]

        # SED
        self.sed_box = widgets.HBox(padding='10px', width="100%")
        self.sed_select = widgets.Dropdown(
            description="SED type: ",
            options=['flat (in Fnu)', 'phoenix', 'blackbody', 'extragalactic', 'star', 'power-law'],
            value='flat (in Fnu)'
        )
        self.bb_temp = widgets.BoundedFloatText(description="Temp (K): ", min=0.0, max=99999.0, step=10, value=6500.0, width=75)
        self.phoenix = widgets.Dropdown(options=sorted(PHOENIX_CONFIG))
        self.galaxies = widgets.Dropdown(options=sorted(GAL_CONFIG))
        self.star = widgets.Dropdown(options=sorted(STAR_CONFIG))
        self.pl_index = widgets.FloatText(description="Index: ", value=1.0, width=50)

        # We have to define anything we're going to use from the beginning, so we do that, and then
        # run the helper function that displays the selected (default) item and hides the rest.
        self.on_sed_change(dummy)
        self.sed_box.children = [self.sed_select, self.phoenix, self.pl_index, self.bb_temp, self.star, self.galaxies]
        self.sed_select.observe(self.on_sed_change)

        # Advanced
        # Geometry
        # The selector goes in one box, and all the geometry selections goes in another so that items can be activated/
        # deactivated as needed.
        self.geom_box = widgets.VBox(width="100%")
        self.source_box = widgets.HBox(width="100%")
        #profile_config_file = os.path.join(refdata, 'source', 'config.json')
        #self.profile_config = get_config(profile_config_file)
        self.src_select = widgets.Dropdown(description="Profile: ", options=['Point Source', 'Flat', '2D Gaussian',
                                                                             'Sersic (Effective Radius)', 'Sersic (Scale Radius)',
                                                                             'Power Law'],
                                             value='Point Source')
        self.source_box.children = [self.src_select]
        self.prof_box = widgets.VBox(width="100%")

        style = {'description_width': 'initial'}

        # We have to define anything we're going to use from the beginning, so we do that, and then
        # run the helper function that displays the selected (default) item and its attendant pieces, and hides the rest.
        self.major = widgets.FloatText(description="Semimajor (arcsec): ", value=0.5, step=0.1, style=style)
        self.minor = widgets.FloatText(description="Semiminor (arcsec): ", value=0.25, step=0.1, style=style)
        self.r_core = widgets.FloatText(description="Core Radius (arcsec): ", value=0.005, step=0.01, style=style)
        self.pos_a = widgets.BoundedFloatText(description="Orientation (deg): ", value=0, min=0, max=359.9, style=style)
        self.sersic = widgets.FloatSlider(description="Sersic Index: ", value=0.5, min=0.3, max=4, readout_format='.1f', style=style)
        self.power = widgets.FloatSlider(description="Power Index: ", value=1, min=0.1, max=10, style=style)
        self.norm = widgets.Dropdown(description="Normalize at: ", options=['infinity', 'scale radius', 'center'], style=style)
        self.norm_flat = widgets.Dropdown(description="Normalize at: ", options=['infinity', 'center'], style=style)
        self.prof_box.children = [self.major, self.minor, self.r_core, self.pos_a, self.norm, self.sersic, self.power]
        self.geom_box.children = [self.source_box, self.prof_box]

        # Position
        self.pos_box = widgets.HBox(padding='10px', width="100%")
        self.pos_x = widgets.BoundedFloatText(description="X Position: ", min=-37.5, step=0.01, max=37.5)
        self.pos_y = widgets.BoundedFloatText(description="Y Position: ", min=-37.5, step=0.01, max=37.5)
        self.pos_box.children = [self.pos_x, self.pos_y]

        # Redshift and Extinction
        self.red_box = widgets.VBox(padding='10px', width="100%")
        self.redshift_box = self.aper_box = widgets.HBox(padding='10px', width="100%")
        self.extinction_box = widgets.HBox(padding='10px', width="100%")
        self.redshift = widgets.BoundedFloatText(description="Redshift (z):", min=0.0, max=99.0, value=0.0, step=0.01, width=70)
        self.redshift_box.children = [self.redshift]

        self.extinction_magnitude = widgets.FloatText(description="Extinction:", value=0, step=0.1, style=style)
        self.extinction_bands = widgets.Dropdown(description="in bandpass:", options=EXTINCTION_BANDS)
        self.extinction_unit = widgets.Dropdown(description="Units of:", options=['mag','nh'])
        self.extinction_law = widgets.Dropdown(description="Using:", options=EXTINCTION_LAWS)
        self.extinction_box.children = [self.extinction_magnitude, self.extinction_unit, self.extinction_bands, self.extinction_law]

        self.red_box.children = [self.redshift_box, self.extinction_box]

        self.advanced_options = widgets.Accordion(children=[self.pos_box, self.geom_box, self.red_box])
        self.advanced_options.set_title(0,"Position")
        self.advanced_options.set_title(1,"Geometry")
        self.advanced_options.set_title(2,"Redshift & Extinction")
        self.advanced_options.selected_index = None

        self.advanced_drop = widgets.Accordion(children=[self.advanced_options])
        self.advanced_drop.set_title(0, "POSITION, GEOMETRY, REDSHIFT & EXTINCTION")
        self.advanced_drop.selected_index = None

        self.grid = GridspecLayout(1,5)
        self.grid[0,4] = widgets.Button(description="Remove Source", disabled=False, button_style="", tooltip="Remove this source from the scene", icon="minus-square")
        self.grid[0,4].on_click(self.hide_source)

        self.container.children = [self.flux_box, self.sed_box, self.advanced_drop, self.grid]

        self.on_prof_change(dummy)
        self.src_select.observe(self.on_prof_change)

    def on_sed_change(self, change):
        # For each possible setting of sed, flip the individual dropdown box sets on and off accordingly.
        self.bb_temp.layout.display = 'none'
        self.phoenix.layout.display = 'none'
        self.galaxies.layout.display = 'none'
        self.star.layout.display = 'none'
        self.pl_index.layout.display = 'none'
        if self.sed_select.value == "phoenix":
            self.phoenix.layout.display = "inline"
        elif self.sed_select.value == "blackbody":
            self.bb_temp.layout.display = "inline"
        elif self.sed_select.value == "extragalactic":
            self.galaxies.layout.display = "inline"
        elif self.sed_select.value == "star":
            self.star.layout.display = "inline"
        elif self.sed_select.value == "power-law":
            self.pl_index.layout.display = "inline"

    def on_prof_change(self,change):
        # For each possible setting of source, flip the properties boxes on and off accordingly.
        self.major.layout.display = 'none'
        self.minor.layout.display = 'none'
        self.pos_a.layout.display = 'none'
        self.sersic.layout.display = 'none'
        self.power.layout.display = 'none'
        self.r_core.layout.display = 'none'
        self.norm.layout.display = 'none'
        self.norm_flat.layout.display = 'none'
        if self.src_select.value == "2D Gaussian":
            self.major.layout.display = 'inline'
            self.minor.layout.display = 'inline'
            self.pos_a.layout.display = 'inline'
            self.norm.layout.display = 'inline'
        elif self.src_select.value == "Flat":
            self.major.layout.display = 'inline'
            self.minor.layout.display = 'inline'
            self.pos_a.layout.display = 'inline'
            self.norm_flat.layout.display = 'inline'
        elif self.src_select.value == "Sersic (Scale Radius)":
            self.major.layout.display = 'inline'
            self.minor.layout.display = 'inline'
            self.pos_a.layout.display = 'inline'
            self.sersic.layout.display = 'inline'
            self.norm.layout.display = 'inline'
        elif self.src_select.value == "Sersic (Effective Radius)":
            self.major.layout.display = 'inline'
            self.minor.layout.display = 'inline'
            self.pos_a.layout.display = 'inline'
            self.sersic.layout.display = 'inline'
            self.norm.layout.display = 'inline'
        elif self.src_select.value == "Power Law":
            self.power.layout.display = 'inline'
            self.r_core.layout.display = 'inline'

    def show_source(self):
        self.active = True
        self.container.layout.display = 'inline'

    def hide_source(self, button):
        self.active = False
        self.container.layout.display = 'none'


class InstObj(object):
    """
    An instrument/detector definition block. It's comprised of two main blocks:
    Filter and Disperser
    Advanced
    """
    def __init__(self,instrument, mode):
        self.container = widgets.VBox(width="100%", background_color="#CCCCCC")

        self.instrument = instrument
        self.instrument.set_mode(mode)
        self.instrument.get_apertures()

        self.aper_box = widgets.HBox(padding='10px', width="100%")
        self.aperture = widgets.Dropdown(description="Aperture:", options=self.instrument.apertures)
        self.aper_box.children = [self.aperture]
        self.on_aperture_change(dummy)

        self.inst_box = widgets.HBox(padding='10px', width="100%")
        self.filt = widgets.Dropdown(description="Filter:", options=self.instrument.filters)
        self.disp = widgets.Dropdown(description="Disperser:", options=self.instrument.dispersers)
        self.inst_box.children = [self.filt, self.disp]

        self.det_box = widgets.HBox(padding='10px', width="100%")
        self.ngroups = widgets.BoundedIntText(description="Groups: ", min=3, max=999, value=6, width=30)
        self.nints = widgets.BoundedIntText(description="Integrations: ", min=1, max=999, value=1, width=30)
        self.nexps = widgets.BoundedIntText(description="Exposures: ", min=1, max=999, value=1, width=30)
        self.det_box.children = [self.ngroups, self.nints, self.nexps]

        self.advanced = widgets.VBox(width="100%", background_color="#CCCCCC")
        self.readout_pattern = widgets.Dropdown(description="Readout Pattern:", options=self.instrument.readout_patterns, value='medium8', width=70)
        self.subarray = widgets.Dropdown(description="Sub-array:", options=self.instrument.subarrays)
        self.advanced.children = [self.readout_pattern, self.subarray]

        self.advanced_drop = widgets.Accordion(children=[self.advanced])
        self.advanced_drop.set_title(0, "ADVANCED")
        self.advanced_drop.selected_index = None
        self.container.children = [self.aper_box, self.inst_box, self.det_box, self.advanced_drop]

        self.aperture.observe(self.on_aperture_change)

    def on_aperture_change(self, change):
        self.instrument.set_aperture(self.aperture.value)


class ImagingApPhotObj(object):
    """
    A strategy definition block. It's comprised of two main blocks:
    Strategy
    Advanced
    """
    def __init__(self):
        self.container = widgets.VBox(width="100%", background_color="#AAAAAA")
        strat_lab = widgets.HTML(value="<b>Imaging Aperture Photometry</b>", margin='5px')

        self.target_box = widgets.HBox(padding='10px', width="100%")
        targ_lab = widgets.HTML(value="Extraction Target: ", margin='5px')
        self.target_x = widgets.BoundedFloatText(description="X:", min=-37.5, max=37.5, value=0, step=0.01, width=30)
        self.target_y = widgets.BoundedFloatText(description="Y:",min=-37.5, max=37.5, value=0, step=0.01, width=30)
        self.target_box.children = [targ_lab, self.target_x,self.target_y]

        self.advanced = widgets.VBox(width="100%", background_color="#AAAAAA")
        self.aperture_box = widgets.HBox(padding='10px', width="100%")
        ap_lab = widgets.HTML(value="Aperture radius (arcsec): ", margin='5px')
        self.ap_size = widgets.BoundedFloatText(min=0.0, max=999.0, value=0.1, width=30)
        self.ap_size.on_trait_change(self.check_ann, 'value')
        self.aperture_box.children = [ap_lab, self.ap_size]

        self.background_box = widgets.VBox(width="100%", background_color="#AAAAAA")
        bg_lab = widgets.HTML(value="Background annulus radii (arcsec): ", margin='5px')
        self.ann_inner = widgets.BoundedFloatText(description="inner", min=0.0, max=999.0, value=0.2, width=30, step=0.01)
        self.ann_inner.on_trait_change(self.check_ann_inner, 'value')
        self.ann_outer = widgets.BoundedFloatText(description="outer", min=0.0, max=999.0, value=0.3, width=30, step=0.01)
        self.ann_outer.on_trait_change(self.check_ann_outer, 'value')
        self.background_box.children = [bg_lab, self.ann_inner, self.ann_outer]
        self.advanced.children = [self.aperture_box, self.background_box]

        self.advanced_drop = widgets.Accordion(children=[self.advanced])
        self.advanced_drop.set_title(0, "ADVANCED")
        self.advanced_drop.selected_index = None
        self.container.children = [strat_lab, self.target_box, self.advanced_drop]

    def check_ann(self, name, value):
        """
        check the background estimation annulus to make sure it's valid

        Parameters
        ----------
        name: string
            not used, but expected for on_trait_change callbacks
        """
        self.check_ann_inner(name=name, value=value)
        self.check_ann_outer(name=name, value=value)

    def check_ann_inner(self, name, value):
        if self.ann_inner.value <= self.ap_size.value:
            self.ann_inner.value = round(self.ap_size.value + 0.1, 3)
        self.check_ann_outer(name=name, value=value)

    def check_ann_outer(self, name, value):
        if self.ann_outer.value - self.ann_inner.value <= 0.1:
            self.ann_outer.value = round(self.ann_inner.value + 0.1, 3)

    def on_src_select(self, name, value):
        if value == 'point':
            self.src_form.visible = False
        else:
            self.src_form.visible = True


class SpecApPhotObj(ImagingApPhotObj):
    """
    A strategy definition block. It's comprised of two main blocks:
    Strategy
    Advanced
    """

    def __init__(self):
        style = {'description_width': 'initial'}
        self.container = widgets.VBox(width="100%", background_color="#AAAAAA")
        strat_lab = widgets.HTML(value="<b>Spectroscopic Aperture Extraction</b>", margin='5px')

        self.target_box = widgets.HBox(padding='10px', width="100%")
        targ_lab = widgets.HTML(value="Extraction Target: ", margin='5px')
        self.target_x = widgets.BoundedFloatText(description="X:", min=-37.5, max=37.5, value=0, step=0.01, width=30)
        self.target_y = widgets.BoundedFloatText(description="Y:", min=-37.5, max=37.5, value=0, step=0.01, width=30)
        self.target_box.children = [targ_lab, self.target_x, self.target_y]

        self.reference_wavelength = widgets.BoundedFloatText(description="Wavelength of Interest", min=0.95, max=1.8, value=1.3, width=30, step=0.01, style=style)

        self.advanced = widgets.VBox(width="100%", background_color="#AAAAAA")
        self.aperture_box = widgets.HBox(padding='10px', width="100%")
        ap_lab = widgets.HTML(value="Aperture half-height (arcsec): ", margin='5px')
        self.ap_size = widgets.BoundedFloatText(min=0.0, max=999.0, value=0.1, width=30)
        self.ap_size.on_trait_change(self.check_ann, 'value')
        self.aperture_box.children = [ap_lab, self.ap_size]

        self.background_box = widgets.VBox(width="100%", background_color="#AAAAAA")
        bg_lab = widgets.HTML(value="Sky Sample Region (arcsec): ", margin='5px')
        self.ann_inner = widgets.BoundedFloatText(description="inner", min=0.0, max=999.0, value=0.2, width=30, step=0.01)
        self.ann_inner.on_trait_change(self.check_ann_inner, 'value')
        self.ann_outer = widgets.BoundedFloatText(description="outer", min=0.0, max=999.0, value=0.35, width=30, step=0.01)
        self.ann_outer.on_trait_change(self.check_ann_outer, 'value')
        self.background_box.children = [bg_lab, self.ann_inner, self.ann_outer]
        self.advanced.children = [self.aperture_box, self.background_box]

        self.advanced_drop = widgets.Accordion(children=[self.advanced])
        self.advanced_drop.set_title(0, "ADVANCED")
        self.advanced_drop.selected_index = None
        self.container.children = [strat_lab, self.target_box, self.reference_wavelength, self.advanced_drop]


class Roman_gui(object):
    """
    create a basic GUI for Roman calculations using ipython widgets. to be run from within a jupyter notebook.
    """
    def __init__(self):
        self.r = {}
        usernotes = widgets.HTML(value="<b>If no advanced settings are changed,</b> a centered point source with no "
                                      "redshift will be computed using readout pattern 'medium8' and subarray '1024x1024', "
                                       "with extraction aperture centered at 0,0 with a size of 0.1 arcsec, and a sky "
                                      "annulus from 0.2-0.3 arcsec."
                                       "<p>The readout patterns are currently inherited from JWST NIRCam. To best "
                                       "approximate readouts as currently envisioned for Roman, we suggest using the "
                                       "following:</p>"
                                       "<ul>"
                                       "<li>Subarray '1024x1024', which has a similar frame time (2.7s) as that "
                                       "expected for the Roman WFI. </li>"
                                       #"<li>Roman does not currently anticipate resetting the detector during an "
                                       #"exposure.  Thus Integrations should be set to 1.</li>"
                                       "<li><i>To approximate the current design for the High Latitude Imaging "
                                       "Survey:</i> readout pattern 'medium8' with Ngroups = 6 </li>"
                                       "<li><i>To approximate the current design for the High Latitude Spectroscopic "
                                       "Survey:</i> readout pattern 'medium8' with Ngroups = 13</li> "
                                       "<li><i>To approximate the current design for the Exoplanet Microlensing "
                                       "Survey:</i> readout pattern 'shallow2' with Ngroups = 4</li")


        self.form = widgets.VBox(width="100%", background_color="#EEE")
        sourcetitle = widgets.HTML(value="<h3>SOURCES</h3>")

        self.source_form = widgets.VBox(width="100%", background_color="#EEEEEE")
        self.sources = []
        for x in range(MAX_SOURCES):
            self.sources.append(SourceObj())
        # Make the first source active
        self.sources[0].show_source()
        self.add_source = widgets.Button(description="Add Source", disabled=False, button_style="", tooltip="Add a new source to the scene", icon="plus-square")
        self.add_source.on_click(self.activate_next)
        source_children = [source.container for source in self.sources]
        source_children.append(self.add_source)
        self.source_form.children = tuple(source_children)
        self.source_form.layout.border = 'solid'

        linebreak = widgets.HTML(value="<br>")
        insttitle = widgets.HTML(value="<h3>INSTRUMENT AND DETECTOR</h3>")
        self.inst_form = widgets.HBox(padding='10px', width="100%")
        self.inst_select = widgets.Dropdown(description="Mode: ", options=['Wide Field Imager'], value='Wide Field Imager')
        self.mode_select = widgets.Dropdown(description="Mode: ", options=['Imaging', 'Spectroscopy'], value='Imaging')
        self.instrument = Instrument('wfi','roman')
        self.instrument.get_modes()
        self.inst_form.children = [self.inst_select, self.mode_select]

        # We have to define both instruments and just hide one of them until selected; we also have to do that with the
        # strategies, and force mode, strategy list, and strategy to change when mode_select changes.
        self.mode_form = widgets.VBox(padding='10px', width="100%")
        self.mode_config = {}
        self.strat_select = {}
        mode_children = []
        mode_strat_children = []
        for mode in self.instrument.modes:
            self.instrument.set_mode(mode)
            self.mode_config[mode] = InstObj(self.instrument, mode)
            mode_children.append(self.mode_config[mode].container)
            self.strat_select[mode] = widgets.Dropdown(description="Strategy: ", options=self.instrument.strategies,
                                                 value=self.instrument.strategies[0])
            mode_strat_children.append(self.strat_select[mode])
        self.mode_form.children = tuple(mode_children)
        self.mode_form.layout.border = 'solid'
        self.mode_select.observe(self.on_mode_change)

        # create both strategies, and arrange to hide all but the one currently selected by the strat_select box
        # (of which there is one for each mode, itself selected by the value of mode_select)
        strattitle = widgets.HTML(value="<h3>EXTRACTION STRATEGY</h3>")
        self.strat_box = widgets.HBox(padding='10px', width="100%")
        self.strat_box.children = tuple(mode_strat_children)
        self.strat_form = widgets.VBox(padding='10px')
        self.strat_config = {}
        self.strat_config['imagingapphot'] = ImagingApPhotObj()
        self.strat_config['specapphot'] = SpecApPhotObj()
        strat_children = [self.strat_config['imagingapphot'].container, self.strat_config['specapphot'].container]
        self.strat_form.children = tuple(strat_children)
        self.strat_form.layout.border = 'solid'

        self.calc_button = widgets.Button(description='Calculate', width="100%", background_color="#bee2c4")
        self.calc_button.on_click(self.run_calc)

        self.result_form = widgets.VBox(padding='10px')
        self.result_form.layout.display = 'none'

        # Make all the boxes, add them to tabs. They'll be (re)populated and displayed when the calculation completes.
        self.plot2d_snr = widgets.Output()
        self.plot2d_detector = widgets.Output()
        self.plot2d_saturation = widgets.Output()
        self.plot2d_ngroups_map = widgets.Output()

        self.plot1d_flux = widgets.Output()
        self.plot1d_bg_only = widgets.Output()
        self.plot1d_sn = widgets.Output()

        self.plot2d_form = widgets.Tab(children=[self.plot2d_snr, self.plot2d_detector, self.plot2d_saturation,
                                                 self.plot2d_ngroups_map])
        self.plot2d_form.set_title(0,'SNR')
        self.plot2d_form.set_title(1,'Detector')
        self.plot2d_form.set_title(2,'Saturation')
        self.plot2d_form.set_title(3,'Groups Before Saturation')

        self.plot1d_form = widgets.Tab(children=[self.plot1d_flux, self.plot1d_bg_only, self.plot1d_sn])
        self.plot1d_form.set_title(0,'Flux')
        self.plot1d_form.set_title(1,'Background Flux Only')
        self.plot1d_form.set_title(2,'SNR')

        self.json_output = widgets.Textarea(value='', description='')
        self.json_box = widgets.Accordion(children=[self.json_output])
        self.json_box.set_title(0,"Copy of Input Dictionary (JSON format)")
        self.json_box.selected_index = None

        self.reportpane = widgets.HTML(value="<hr>", margin='5px')

        self.computing_notice = widgets.HTML(value="<i>Calculating... please wait.</i>", margin='5px')
        self.computing_notice.layout.display = 'none'

        self.tab_form = widgets.HBox(padding='10px', width="100%", pack='center')
        self.tab_form.children = [self.reportpane]

        self.result_form.children = [self.plot2d_form, self.plot1d_form, self.tab_form, self.json_box]

        self.form.children = [
            usernotes,
            sourcetitle,
            self.source_form,
            linebreak,
            insttitle,
            self.inst_form,
            self.mode_form,
            linebreak,
            strattitle,
            self.strat_box,
            self.strat_form,
            linebreak,
            self.calc_button,
            self.computing_notice,
            self.result_form
        ]

        self.on_mode_change(dummy)
        self.on_strat_change(dummy)

    def activate_next(self, button):
        next_source = [x for x in range(len(self.sources)) if self.sources[x].active == False]
        if len(next_source) > 0:
            #self.sources[next_source[0]] = SourceObj()
            self.sources[next_source[0]].show_source()

    def plot2d(self, plotname, unitstring):
        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        t = self.r['transform']
        xmin = t['x_min']
        xmax = t['x_max']
        ymin = t['y_min']
        ymax = t['y_max']
        extent = [xmin, xmax, ymin, ymax]
        im = ax1.imshow(self.r['2d'][plotname],cmap='coolwarm', extent=extent)
        ax1.set_xlabel('arcsec')
        ax1.set_ylabel('arcsec')
        orient = "vertical"
        if self.r['2d'][plotname].shape[0] != self.r['2d'][plotname].shape[1]:
            orient = "horizontal"
        if plotname == "saturation":
            norm = matplotlib.colors.Normalize(vmin=0, vmax=2)
            im.set_norm(norm)
            c = fig.colorbar(im, ax=ax1, orientation=orient, label=unitstring, ticks=[0, 1, 2])
            c.ax.set_xticklabels(["None", "Partial", "Full"])
        else:
            c = fig.colorbar(im, ax=ax1, orientation=orient, label=unitstring)
        plt.tight_layout()
        plt.show()

    def plot1d(self, plotname, unitstring):
        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        # single points (imaging, for example) need to be scatter-plotted.
        if len(self.r['1d'][plotname][0]) == 1:
            im = ax1.scatter(self.r['1d'][plotname][0], self.r['1d'][plotname][1])
        else:
            im = ax1.plot(self.r['1d'][plotname][0], self.r['1d'][plotname][1])
        ax1.set_ylabel(unitstring)
        ax1.set_xlabel('microns')
        plt.tight_layout()
        plt.show()

    def on_mode_change(self,change):
        for mode in self.mode_config:
            self.mode_config[mode].container.layout.display = 'none'
            self.strat_select[mode].layout.display = 'none'
        self.mode_config[self.mode_select.value.lower()].container.layout.display = 'inline'
        self.strat_select[self.mode_select.value.lower()].layout.display = 'inline'
        self.on_strat_change(dummy)

    def on_strat_change(self,change):
        for strat in self.strat_config:
            self.strat_config[strat].container.layout.display = 'none'
        self.strat_config[self.strat_select[self.mode_select.value.lower()].value.lower()].container.layout.display = 'inline'

    def update_plots(self):
        """
        update the 1D and 2D plots.  they're part of the same figure so have to be drawn together.  hard to do
        two independent plots in one cell in a notebook.
        """
        with self.plot2d_snr:
            self.plot2d('snr', 'S/N')
            clr = clear_output(wait=True)

        with self.plot2d_detector:
            self.plot2d('detector', 'electrons / s')
            clr = clear_output(wait=True)

        with self.plot2d_saturation:
            self.plot2d('saturation', ' ')
            clr = clear_output(wait=True)

        with self.plot2d_ngroups_map:
            self.plot2d('ngroups_map', '#')
            clr = clear_output(wait=True)

        with self.plot1d_flux:
            self.plot1d('extracted_flux', 'electrons / s')
            clr = clear_output(wait=True)

        with self.plot1d_bg_only:
            self.plot1d('extracted_bg_only', 'electrons / s')
            clr = clear_output(wait=True)

        with self.plot1d_sn:
            self.plot1d('sn', '#')
            clr = clear_output(wait=True)


    @property
    def display(self):
        """
        display the GUI
        """
        display(self.form)

    @property
    def calc_results(self):
        """
        return the calculation results
        """
        return self.r

    @calc_results.setter
    def calc_results(self, r):
        self.r = r
        self.update_plots()

    def on_units_select(self, name, value):
        if value == 'abmag':
            self.flux.value = 25.0
        else:
            self.flux.value = 1.0

    def run_calc(self, b):
        # notify the user that we're calculating.
        self.computing_notice.layout.display = 'inline'
        calc_mode = self.mode_select.value.lower()
        calc_strat = self.strat_select[calc_mode].value.lower()
        c = build_default_calc("roman", "wfi", calc_mode, method=calc_strat)
        c['configuration']['detector']['nexp'] = self.mode_config[calc_mode].nexps.value
        c['configuration']['detector']['ngroup'] = self.mode_config[calc_mode].ngroups.value
        c['configuration']['detector']['nint'] = self.mode_config[calc_mode].nints.value
        c['configuration']['detector']['readout_pattern'] = self.mode_config[calc_mode].readout_pattern.value
        c['configuration']['detector']['subarray'] = self.mode_config[calc_mode].subarray.value
        c['configuration']['instrument']['filter'] = self.mode_config[calc_mode].filt.value
        c['configuration']['instrument']['disperser'] = self.mode_config[calc_mode].disp.value

        c['scene'] = []

        for source in self.sources:
            if source.active:
                s = build_default_source(geometry=geom_mapping[source.src_select.value])
                if source.src_select.value == "Power":
                    s['shape']['r_core'] = source.r_core.value
                    s['shape']['power_index'] = source.power.value
                elif source.src_select.value == "Flat":
                    s['shape']['major'] = source.major.value
                    s['shape']['minor'] = source.minor.value
                    s['shape']['norm_method'] = norm_mapping[source.norm_flat.value]
                    s['position']['orientation'] = source.pos_a.value
                elif source.src_select.value == "2D Gaussian":
                    s['shape']['major'] = source.major.value
                    s['shape']['minor'] = source.minor.value
                    s['shape']['norm_method'] = norm_mapping[source.norm.value]
                    s['position']['orientation'] = source.pos_a.value
                elif source.src_select.value == "Sersic (Effective Radius)":
                    s['shape']['major'] = source.major.value
                    s['shape']['minor'] = source.minor.value
                    s['shape']['norm_method'] = norm_mapping[source.norm.value]
                    s['shape']['sersic_index'] = source.sersic.value
                    s['position']['orientation'] = source.pos_a.value
                elif source.src_select.value == "Sersic (Scale Radius)":
                    s['shape']['major'] = source.major.value
                    s['shape']['minor'] = source.minor.value
                    s['shape']['norm_method'] = norm_mapping[source.norm.value]
                    s['shape']['sersic_index'] = source.sersic.value
                    s['position']['orientation'] = source.pos_a.value
                else:
                    pass

                s['position']['x_offset'] = source.pos_x.value
                s['position']['y_offset'] = source.pos_y.value

                s['spectrum']['redshift'] = source.redshift.value
                s['spectrum']['extinction']['law'] = extinction_mapping[source.extinction_law.value]
                s['spectrum']['extinction']['value'] = source.extinction_magnitude.value
                s['spectrum']['extinction']['bandpass'] = source.extinction_bands.value
                s['spectrum']['extinction']['unit'] = source.extinction_unit.value
                s['spectrum']['normalization']['norm_flux'] = source.flux.value
                s['spectrum']['normalization']['norm_fluxunit'] = source.funits.value
                s['spectrum']['normalization']['norm_wave'] = source.wave.value

                sed = source.sed_select.value
                # the source is always rebuilt from scratch, so it always starts out as a flat source
                if sed == "power-law":
                    s['spectrum']['sed']['sed_type'] = "powerlaw"
                    s['spectrum']['sed']['index'] = source.pl_index.value
                elif sed == "blackbody":
                    s['spectrum']['sed']['sed_type'] = "blackbody"
                    s['spectrum']['sed']['temp'] = source.bb_temp.value
                    del s['spectrum']['sed']['unit']
                elif sed == "phoenix":
                    s['spectrum']['sed']['sed_type'] = "phoenix"
                    s['spectrum']['sed']['key'] = source.phoenix_config[source.phoenix.value]
                    del s['spectrum']['sed']['unit']
                elif sed == "extragalactic":
                    s['spectrum']['sed']['sed_type'] = "brown"
                    s['spectrum']['sed']['key'] = source.gal_config[source.galaxies.value]
                    del s['spectrum']['sed']['unit']
                elif sed == "star":
                    s['spectrum']['sed']['sed_type'] = "hst_calspec"
                    s['spectrum']['sed']['key'] = source.star_config[source.star.value]
                    del s['spectrum']['sed']['unit']

                c['scene'].append(s)

        c['strategy']['aperture_size'] = self.strat_config[calc_strat].ap_size.value
        ann = [self.strat_config[calc_strat].ann_inner.value, self.strat_config[calc_strat].ann_outer.value]
        c['strategy']['sky_annulus'] = ann
        if calc_strat == 'specapphot':
            c['strategy']['reference_wavelength'] = self.strat_config[calc_strat].reference_wavelength.value

        self.r = perform_calculation(c, dict_report=True)
        self.calc_input = c

        resultpane = ["<hr><h3><center>RESULTS</center></h3><hr><table>"]

        # print the scalar products with what (I think) are the correct units.
        for x in sorted(self.r["scalar"]):
            resultitem = self.r["scalar"][x]
            if type(resultitem) in [float, np.float, np.float32, np.float64]:
                resultitem = "{:.3f}".format(resultitem)
            if "time" in x:
                suffix = " sec"
            elif "size" in x or "offset" in x:
                suffix = " arcsec"
            elif "area" in x:  # checking this before background means the
                            # background_area will be given the correct units.
                suffix = " pixel^2"
            elif "wavelength" in x:
                suffix = " microns"
            elif (("extracted" in x) or ("sky" in x)) or ("total" in x):
                suffix = " e-/sec"
            elif "background" in x:
                suffix = " MJy/sr"
            else:
                suffix = ""
            resultpane.append("<tr><td><b>{}</b></td><td align='right'>{}</td><td>&nbsp;&nbsp;{}</td></tr>".format(x.replace('_', ' ').title(), resultitem, suffix))
        resultpane.append('</table>')
        resultpane.append('<table>')
        resultpane.append("<tr><td><b><i>WARNINGS</i></b></td></tr>")
        for x in self.r['warnings']:
            resultpane.append("<tr><td></td><td></td><td>{}</td></tr>".format(self.r['warnings'][x]))
        resultpane.append('</table>')


        self.reportpane.value = ''.join(resultpane)

        self.json_output.value = json.dumps(self.calc_input, sort_keys=True, indent=4, cls=NumPyArangeEncoder, separators=(',', ': '))

        self.update_plots()
        # Now set the result form to be shown
        self.computing_notice.layout.display = 'none'
        self.result_form.layout.display = 'inline'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, absolute_import

import os
import six
from functools import reduce
from warnings import warn
import numpy as np
import scipy.sparse as sparse
import operator

import astropy.io.fits as fits
from astropy.convolution import Gaussian1DKernel, convolve

from . import exposure as exp
from . import config as cf
from .psf_library import PSFLibrary
from .io_utils import read_json, ref_data_interp, ref_data_column
from .utils import merge_data, spectrum_resample
from .telescope import TelescopeConfig
from .custom_exceptions import EngineInputError, DataError, UnsupportedError, InternalError, DataConfigurationError
from .constants import DEFAULT_DISPERSION

default_refdata_directory = cf.default_refdata_directory

def notin(i,l):
    return i not in l

class InstrumentConfig(TelescopeConfig):

    """
    This is a trivial top-level class for now. It overrides TelescopeConfig._get_config() to get the configuration
    information from a different place, currently under $pandeia_refdata/<tel_name>/<inst_name>
    """

    def _get_config(self):
        """
        Read default source configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        self.inst_name = self.__class__.__name__.lower()
        self.ref_dir = os.path.join(default_refdata_directory, self.telescope.tel_name, self.inst_name)
        config = read_json(os.path.join(self.ref_dir, "config.json"), raise_except=True)

        # Grab the warning list, if it exists, in the config file.
        self.warning_list = []
        if 'warning_list' in config:
            self.warning_list = config['warning_list']

        # pop the defaults entry out
        if "defaults" in config:
            defaults = config.pop("defaults")
            if self.mode is None:
                # if no mode specified, then get default mode
                if "mode" in defaults:
                    self.mode = defaults.pop("mode")
                else:
                    message = "No mode or default mode specified."
                    raise DataConfigurationError(value=message)
            # make sure mode is valid and if so use it to key the values to put into config
            if self.mode in defaults:
                config.update(defaults[self.mode])
            else:
                message = "Invalid %s mode: %s" % (self.inst_name, self.mode)
                raise DataConfigurationError(value=message)
        return config


class Instrument(InstrumentConfig):

    """
    This is the primary Instrument class that sets up the configuration and provides methods common
    to all instruments.

    Parameters
    ----------
    telescope: Telescope sub-class instance
        The telescope that contains the instrument
    mode: str
        Desired instrument mode
    config: dict
        dictionary containing necessary configuration information
    **kwargs: list of keyword/value pairs
        parameter keyword and value pairs to augment defaults and config
    """

    def __init__(self, telescope=None, mode=None, config={}, webapp=False, **kwargs):
        self.webapp = webapp
        if telescope is None:
            msg = "Telescope not defined for %s!" % self.__class__.__name__
            raise EngineInputError(value=msg)
        if mode is None:
            msg = "Instrument mode not defined for %s!" % self.__class__.__name__
            raise EngineInputError(value=msg)
        self.telescope = telescope
        self.mode = mode
        self.order = None
        self.detector_key = 'detector' # default value

        InstrumentConfig.__init__(self, config, webapp=webapp, **kwargs)

        # make sure telescope is configured properly. if not, the reference data directory
        # is likely not defined or doesn't exist.
        try:
            # make sure instrument is supported
            if self.inst_name not in self.telescope.instruments:
                message = "Unsupported %s instrument: %s" % (self.telescope.tel_name, self.inst_name)
                raise EngineInputError(value=message)
        except AttributeError as e:
            message = "Telescope/instrument configuration data missing.  "
            message += "Reference data directory possibly missing or not defined: %s" % default_refdata_directory
            message += "Original exception: %s" % e
            raise InternalError(value=message)

        # make sure required parameters are defined and do strict API checking if webapp is true
        all_config = merge_data(config, dict(**kwargs))
        self._nested_api_checks(all_config, webapp=webapp)

        self.exposure_spec = self.get_exposure_pars()
        self._loadpsfs()

        # Check the instrument and strategy for the calculation against
        # the configured warnings in the instrument's config.json file.
        #
        # There are times that an Instrument instance will be created in order to just read
        # some values (e.g., bandpass in normalization.py's function _get_bandpass()) and in
        # these cases the only thing passed in is the instrument information. So, we are going
        # to check the warnings only if 'detector' and 'instrument' are passed in through all_config.
        if 'detector' in all_config and 'instrument' in all_config:
            self.check_warnings(all_config)

    def _nested_api_checks(self, conf, webapp=False):
        for section in self.instrument_pars:
            for p in self.instrument_pars[section]:
                # if parameter is not already set, then it's a DataError since config file should have set it to a default.
                if p not in getattr(self, section):
                    msg = "Missing required %s parameter %s from default configuration for %s." % (
                        section,
                        p,
                        self.__class__.__name__
                    )
                    raise DataError(value=msg)
                # if we're doing API checking, make sure everything we need is specified by the inputs.
                if webapp:
                    if p not in conf[section]:
                        key = "%s_missing_%s_%s" % (
                            self.__class__.__name__.lower(),
                            section,
                            p
                        )
                        self.warnings[key] = "Missing %s API parameter %s for %s. Using default value of %s." % (
                            section,
                            p,
                            self.__class__.__name__,
                            getattr(self, section)[p]
                        )

    def check_warnings(self, input_dict):
        """

        :param warnings_list: list of warnings to check for and add to the warnings list
        :param input_dict: input as passed into the function
        :return:
        """

        # Each thing we want to be able to get the value of in the input dictionary
        # should be listed here. The "value" part is a list of nested keys to get the actual value.
        value_lookup = {
            'instrument':       ['instrument', 'instrument'],
            'mode':             ['instrument', 'mode'],
            'filter':           ['instrument', 'filter'],
            'disperser':        ['instrument', 'disperser'],
            'aperture':         ['instrument', 'aperture'],
            'nexp':             ['detector', 'nexp'],
            'ngroup':           ['detector', 'ngroup'],
            'nint':             ['detector', 'nint'],
            'max_total_groups': ['detector', 'max_total_groups'],
            'readout_pattern':  ['detector', 'readout_pattern'],
            'subarray':         ['detector', 'subarray'],
        }

        # Run through each warning in the warnings_list to see if we need to return it.
        for warning in self.warning_list:

            # Run through all the parameters for this warning to see if all are true
            need_warning = True
            thewarning = warning['warning_text']
            for key, expected_value in list(warning['parameters'].items()):

                # Get the value from the json dict sent to the engine.  This does a lookup
                # based on the value_lookup above to see where in the input json dict we
                # have to look for the value that was input.
                # https://stackoverflow.com/questions/13687924/setting-a-value-in-a-nested-python-dictionary-given-a-list-of-indices-and-value

                operator_text = '=='
                if '__' in key:
                    key, operator_text = key.split('__', 1)

                # https://stackoverflow.com/questions/9320335/accessing-python-dict-with-multiple-key-lookup-string
                input_value = reduce(dict.get, value_lookup[key], input_dict)

                # Setup the comparison operator to use below.
                cmp = '=='
                if operator_text == '==':
                    cmp = operator.eq
                elif operator_text == 'ge':
                    cmp = operator.ge
                elif operator_text == 'le':
                    cmp = operator.le
                elif operator_text == 'gt':
                    cmp = operator.gt
                elif operator_text == 'lt':
                    cmp = operator.lt
                elif operator_text == 'notin':
                    cmp = lambda a,b: a not in b
                else:
                    raise InternalError(value="Unknown operator on key {}__{}".format(key, operator_text))

                # For some reason there started to be collisions on comparison of str vs unicode so I am
                # going to convert things to string that might be string or unicode.  That way the comparison
                # will work properly.  In Python3 this will likely be able to be removed.
                if isinstance(input_value, six.string_types):
                    input_value = str(input_value)
                    if isinstance(expected_value, six.string_types):
                        expected_value = str(expected_value)
                    elif isinstance(expected_value, list):
                        expected_value = [str(x) for x in expected_value]

                # Compare the input_value with the expected value
                if (  # do operator comparison for value
                    (isinstance(expected_value, (six.string_types, six.integer_types, float)) and cmp(input_value, expected_value)) or
                    # do set inclusion or exclusion for list
                        (isinstance(expected_value, list) and
                             ((operator_text == '==' and input_value in expected_value) or
                             (operator_text == 'notin' and not input_value in expected_value))
                        )
                    ):
                    pass  # we do need a warning, based on this test
                else:
                    need_warning = False

                # Replace $mode, $instrument, $filter etc in the output warning with the actual value.
                # special syntax if we don't want it to be uppercase.
                if '$~'+key in thewarning:
                    thewarning = thewarning.replace('$~' + key, input_value)
                if '$'+key in thewarning:
                    thewarning = thewarning.replace('$' + key, input_value.upper())

            if need_warning:
                self.warnings[thewarning] = thewarning

    def _loadpsfs(self):
        """
        Load the PSF library for the current instrument configuration.  In most cases, this is based on the aperture.
        In some cases (e.g. NIRCam coronagraphy with bar-shaped masks), other configuration parameters come into play.
        """
        self.psf_library = self._load_psf_library(self.instrument['aperture'])

    def _load_psf_library(self,psf_key):

        psf_path = os.path.join(self.ref_dir, "psfs")
        library = PSFLibrary(path=psf_path, aperture=psf_key)

        return library

    @property
    def qe_key(self):
        """
        set default qe_key for looking up QE reference data. MIRI different from others and overloads this.
        """
        return 'qe'

    def get_detector_pars(self):
        """
        set default detector_key for looking up the detector that is used with a certain aperture. This can be
        overloaded for specific instruments.
        """
        return self.read_detector_pars()

    def get_name(self):
        """
        Return instrument's name
        """
        return self.inst_name

    def get_aperture(self):
        """
        Return configured aperture
        """
        return self.instrument['aperture']

    def _interp_refdata(self, wave, key, colname, default=None):
        """
        Read in requested reference file and return requested data interpolated onto wave

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array onto which the reference data is to be interpolated
        key: str
            Key to get filename from self.paths
        colname: str
            Column to retrieve from reference file
        default: None or float
            Default value to return in case of missing file or column

        Returns
        -------
        ref: numpy.ndarray or float
            If file exists, return reference_data(wave). Else return default.
        """
        ref = None
        if key is not None and key in self.paths:
            ref_file = os.path.join(self.ref_dir, self.paths[key])
            if ref_file is not None:
                ref = ref_data_interp(ref_file, wave, colname)
        if ref is None and default is None:
            msg = "No reference data found and no default value supplied for %s in %s." % (colname, key)
            raise DataError(value=msg)
        elif ref is None and default is not None:
            ref = np.ones_like(wave)*default
        return ref

    def _get_throughput(self, wave, key):
        """
        Wrap self._interp_refdata for getting throughput data

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array onto which the throughput curve is to be interpolated
        key: str
            Key to get filename from self.paths

        Returns
        -------
        eff: numpy.ndarray or float
            If ref file exists, return efficiency(wave), else return 1.0
        """
        eff = self._interp_refdata(wave, key, colname='throughput', default=1.0)
        return eff

    def _get_disperser(self):
        """
        The same disperser can have different configurations depending on which order is
        being used. Use self.order to build a new key for looking up configuration data for the
        given order.

        Returns
        -------
        key: str
            Key to look up order-specific configuration data
        """
        key = self.instrument['disperser']
        if self.order is not None:
            key = "%s_%d" % (key, self.order)
            key = key.replace('-','m')
        return key

    def _get_dispersion_key(self):
        """
        Different instruments have different ways to figure out which disperser to use. This is also used
        to look up reference data for multiple orders from the same disperser.

        Returns
        -------
        key: str
            Key used to get dispersion file out of self.paths
        """
        disperser = self._get_disperser()
        key = "%s_disp" % disperser
        return key

    def get_filter_eff(self, wave):
        """
        Read in filter throughput curve

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Filter throughput as a function of wave
        """
        eff = self._get_throughput(wave, self.instrument['filter'])
        return eff

    def get_disperser_eff(self, wave):
        """
        Read in efficiency curve for configured grating

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Disperser efficiency as a function of wave
        """
        key = self._get_disperser()
        eff = self._get_throughput(wave, key)
        return eff

    def get_internal_eff(self, wave):
        """
        Read in internal efficiency

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Internal throughput as a function of wave
        """
        eff = self._get_throughput(wave, 'internal')
        return eff

    def get_detector_qe(self, wave):
        """
        Read in detector quantum efficiency. This will likely need to be
        overloaded on a per detector basis.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        qe: numpy.ndarray or float
            Quantum efficiency as a function of wave
        """
        qe = self._get_throughput(wave, self.qe_key)
        return qe

    def get_quantum_yield(self, wave):
        """
        Read in detector quantum yield if it's there. If not, return unity (trivial yield).

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        quantum_yield: numpy.ndarray or float
            Quantum efficiency as a function of wave
        """
        q_yield = self._interp_refdata(wave, self.qe_key, colname='conversion', default=1.0)
        fano_factor = (3.0 * q_yield - q_yield**2. - 2.) / q_yield  # two level model, valid for quantum yield between 1 and ~2.
        return q_yield, fano_factor

    def get_total_eff(self, wave):
        """
        Calculate combined throughput of entire telescope/instrument/detector system

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Total system throughput as a function of wave
        """
        ote_int = self.telescope.get_ote_eff(wave)
        filter_eff = self.get_filter_eff(wave)
        disperser_eff = self.get_disperser_eff(wave)
        internal_eff = self.get_internal_eff(wave)
        det_qe = self.get_detector_qe(wave)

        eff = ote_int * filter_eff * disperser_eff * internal_eff * det_qe
        return eff

    def get_dispersion(self, wave):
        """
        Read in dispersion

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim dispersion data onto

        Returns
        -------
        dispersion: numpy.ndarray or float
            Dispersion as a function of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = self._get_dispersion_key()
            dispersion = self._interp_refdata(wave, key, colname='dlds', default=DEFAULT_DISPERSION)
        else:
            dispersion = 0.0
        return dispersion

    def get_trace(self, wave):
        """
        Implement spectral traces, wavelength-dependent shifts of the spatial axis.  This base implementation only returns
        an array of 0's of the appropriate size.  This is to make the code implementation of this fully general.
        Instruments that need to account for such spatial distortions will need to overload this to load and implement
        the appropriate reference data.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim trace data onto

        Returns
        -------
        trace: numpy.ndarray or float
            Spectral trace offset from center of FOV as a function of wave
        """
        trace = np.zeros_like(wave)

        return trace

    def get_detector_pixels(self, wave):
        """
        Read in detector pixel positions for each wavelength in wave_pix

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim pixel position data onto

        Returns
        -------
        pixels: numpy.ndarray or None
            Detector pixel positions corresponding to each element of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = "%s_wavepix" % disperser
            try:
                pixels = self._interp_refdata(wave, key, colname='pixels', default=np.arange(len(wave)))
            except DataError as e:
                pixels = None
        else:
            pixels = None
        return pixels

    def get_resolving_power(self, wave):
        """
        Read in resolving power

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate resolving power onto

        Returns
        -------
        rpower: numpy.ndarray or None
            Resolving power as a function of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = self._get_dispersion_key()
            rpower = self._interp_refdata(wave, key, colname='R', default=wave/(DEFAULT_DISPERSION*2))
        else:
            rpower = None
        return rpower

    def get_wave_blaze(self):
        """
        Get wavelength vector used in the grating efficiency (blaze) file

        Returns
        -------
        wave_blaze: numpy.ndarray
            Wavelength vector from the grating efficiency file
        """
        key = self._get_disperser()
        blaze_file = os.path.join(self.ref_dir, self.paths[key])
        msg = "Blaze file not found: %s" % blaze_file
        wave_blaze = ref_data_column(blaze_file, colname='wavelength', error_msg=msg)
        return wave_blaze

    def get_wave_filter(self):
        """
        Get wavelength vector used in the filter throughput file

        Returns
        -------
        wave_filter: numpy.ndarray
            Wavelength vector from the filter throughput file
        """
        key = self.instrument['filter']
        filter_file = os.path.join(self.ref_dir, self.paths[key])
        msg = "Filter file not found: %s" % filter_file
        wave_filter = ref_data_column(filter_file, colname='wavelength', error_msg=msg)
        return wave_filter

    def get_wave_pix(self):
        """
        Get wavelength vector to convert pixel position to wavelength

        Returns
        -------
        wavepix: numpy.ndarray
            Wavelength vector mapping pixels to wavelengths
        """
        # usually wavepix is lifted from the dispersion file, but in some cases (e.g. NIRISS SOSS)
        # there is a specific, calibrated mapping of wavelength to pixels. look for that first and then
        # fall back to the dispersion file.
        wavepix_key = "%s_wavepix" % self._get_disperser()
        if wavepix_key in self.paths:
            # if we have a specifically calibrated mapping of wavelengths to pixels (e.g. for SOSS),
            # look up that file and use that mapping.
            wavecal_file = os.path.join(self.ref_dir, self.paths[wavepix_key])
            msg = "Wavelength calibration file not found: %s" % wavecal_file
            wavepix = ref_data_column(wavecal_file, colname='wavelength', error_msg=msg)
        else:

            # in dispersion files, there is no guarantee that the table is sampled in pixel spacing. This information
            # is encoded in the dlds column so calculate it from there.
            key = self._get_dispersion_key()
            if key in self.paths:
                dispersion_file = os.path.join(self.ref_dir, self.paths[key])
                msg = "Dispersion file not found: %s" % dispersion_file
                wave = ref_data_column(dispersion_file, colname='wavelength', error_msg=msg)
                dlds = ref_data_column(dispersion_file, colname='dlds', error_msg=msg)
            else:
                # for spectroscopic order 0, there will be no dispersion file, and thus no reference in the
                # config.json path{} block.
                wave = self.get_wave_blaze()
                dlds = np.ones_like(wave) * DEFAULT_DISPERSION
            pixel = np.cumsum(1.0 / dlds * np.gradient(wave))
            pixel_integer = np.arange(int(pixel[0]), int(pixel[-1]))
            wavepix = np.interp(pixel_integer, pixel, wave)

        return wavepix

    def get_ipc_kernel(self):
        """
        Get inter-pixel capacitance (IPC) kernel

        Returns
        -------
        kernel: numpy.ndarray
            IPC kernel data
        """
        key = "ipc_kernel"
        ipc_file = os.path.join(self.ref_dir, self.paths[key])
        try:
            kernel = fits.getdata(ipc_file)
        except IOError as e:
            msg = "Error reading IPC kernel reference file: %s. " % ipc_file
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataError(value=msg)
        # The new convolution algorithm in scipy 1.4+ can't handle comparing data of different types
        return np.asarray(kernel,dtype=np.float32)

    def read_detector_pars(self):
        """
        Return detector-specific parameters.

        This reads the JSON structure so values can be dependent on aperture, subarray,
        readout pattern, and/or instrument mode.

        Returns
        -------
        det_dict: dict
            A dictionary of the correct detector parameters
        """
        det_dict = {}

        for parameter_key in self.det_pars:
            parameter_value = self.read_config_param(self.det_pars,parameter_key)

            det_dict[parameter_key] = parameter_value

        return det_dict

    def get_variance_fudge(self, wave):
        """
        Some instruments (e.g. MIRI) have measured noise properties that are not well explained
        or modeled by Pandeia's current noise model. This method returns the multiplicative fudge factor
        that is to be applied to the pixel rate variance to match IDT expectations. The fudge factor can be
        chromatic so needs to accept an array of wavelengths to interpolate over. By default this returns
        a 1's array the same length as wave scaled by the 'var_fudge' scalar value defined in det_pars.

        Parameters
        ----------
        wave: 1D numpy.ndarray
            Wavelengths to interpolate chromatic variance over
        """
        det_pars = self.get_detector_pars()
        var_fudge = det_pars.get('var_fudge', 1.0) * np.ones(len(wave))
        return var_fudge

    def get_readnoise_correlation_matrix(self, shape):
        """
        Grab correlated readnoise data out of reference file and build a correlation
        matrix out of it.

        Parameters
        ----------
        shape: list-like

        Returns
        -------
        correlation_matrix: Scipy.sparse.csr_matrix
            Compressed sparse row correlation matrix
        """
        key = "rn_corr"
        correlation_file = os.path.join(self.ref_dir, self.paths[key])
        try:
            correlation = fits.getdata(correlation_file)
        except IOError as e:
            msg = "Error reading RN correlation reference file: %s. " % correlation_file
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataError(value=msg)

        nframe_range = correlation.shape[0]
        nx = correlation.shape[2]
        ny = correlation.shape[1]

        if nx != ny:
            message = 'Only square correlation matrices are supported.'
            raise UnsupportedError(value=message)

        if self.exposure_spec.nframe > nframe_range:
            # nframe longer than full range of correlation so use the full range
            corr_range = nframe_range
        else:
            # nframe shorter than range of correlation so use subset of correlation data
            corr_range = self.exposure_spec.nframe
        correlation_data = sparse.csr_matrix(correlation[corr_range, :, :])

        empty_matrix = sparse.coo_matrix((1024, 1024))
        correlation_matrix = sparse.bmat([[empty_matrix, None, empty_matrix],
                                          [None, correlation_data, None],
                                          [empty_matrix, None, empty_matrix]])
        correlation_matrix = sparse.csr_matrix(correlation_matrix)

        return correlation_matrix

    def get_exposure_pars(self, name="pattern_name"):
        """
        Define exposure parameters from instrument and environment.

        Parameters
        ----------
        name: str
            Optional name to assign ExposureSpecification instance

        Returns
        -------
        exposure_spec: pandeia.engine.exposure.ExposureSpecification instance
            Exposure parameters encapsulated within ExposureSpecification instance
        """
        # These are values that get set when the Instrument is instantiated, either from defaults
        # or passed configuration data.
        readout_pattern = self.detector["readout_pattern"]
        subarray = self.detector["subarray"]
        ngroup = self.detector["ngroup"]
        nexp = self.detector["nexp"]
        nint = self.detector["nint"]

        det_type = self.get_detector_pars()['det_type']

        # These are defined by the Instrument's reference data as Instrument properties.
        tframe = self.subarray_config['default'][subarray]["tframe"]
        tfffr = self.subarray_config['default'][subarray]["tfffr"]
        if readout_pattern in self.subarray_config and subarray in self.subarray_config[readout_pattern]:
            tframe = self.subarray_config[readout_pattern][subarray]["tframe"]
            tfffr = self.subarray_config[readout_pattern][subarray]["tfffr"]
        nframe = self.readout_pattern_config[readout_pattern]["nframe"]
        nskip = self.readout_pattern_config[readout_pattern]["nskip"]

        # If these are trivial, we don't have to define them.
        if "nsample" in self.readout_pattern_config[readout_pattern]:
            nsample = self.readout_pattern_config[readout_pattern]["nsample"]
            nsample_skip = self.readout_pattern_config[readout_pattern]["nsample_skip"]
        else:
            nsample = 1
            nsample_skip = 0

        # Going from general to specific: start with the default reset values, then check the defaults, then check
        # the actual readout pattern for this subarray. Use the most appropriate reset values for this setup, otherwise
        # the defaults
        nreset1 = 1
        nreset2 = 1
        if "nreset1" in self.subarray_config['default'][subarray]:
            nreset1 = self.subarray_config['default'][subarray]["nreset1"]
            nreset2 = self.subarray_config['default'][subarray]["nreset2"]
        if readout_pattern in self.subarray_config and subarray in self.subarray_config[readout_pattern]:
            if "nreset1" in self.subarray_config[readout_pattern][subarray]:
                nreset1 = self.subarray_config[readout_pattern][subarray]["nreset1"]
                nreset2 = self.subarray_config[readout_pattern][subarray]["nreset2"]


        # These are defined in the reference data as "ramp_config" and apply to a whole instrument
        nprerej = self.read_config_param(self.ramp_config, "nprerej")
        npostrej = self.read_config_param(self.ramp_config, "npostrej")

        # We allow alternate ways to derive a signal from a detector slope,
        # but multiaccum is the default if no mode is provided. Slopemode
        # must be matched with a corresponding ExposureSpecification class.

        slopemode = self.detector.get("slopemode", "multiaccum")

        if slopemode == "multiaccum":
            exposure_spec = exp.ExposureSpecification(
                name,
                ngroup,
                nint,
                nexp,
                tframe,
                nframe=nframe,
                subarray=subarray,
                nskip=nskip,
                nsample=nsample,
                nsample_skip=nsample_skip,
                nprerej=nprerej,
                npostrej=npostrej,
                nreset1=nreset1,
                nreset2=nreset2,
                det_type=det_type,
                tfffr=tfffr
            )
        elif slopemode == "target_acq":
            ngroup_extract = self.detector["ngroup_extract"]
            exposure_spec = exp.ExposureSpecificationTA(
                name,
                ngroup,
                ngroup_extract,
                nint,
                nexp,
                tframe,
                nframe=nframe,
                subarray=subarray,
                nskip=nskip,
                nsample=nsample,
                nsample_skip=nsample_skip,
                nprerej=nprerej,
                npostrej=npostrej,
                nreset1=nreset1,
                nreset2=nreset2,
                det_type=det_type,
                tfffr=tfffr
            )
        else:
            msg = "%s is not a valid slope mode." % (slopemode)
            raise DataConfigurationError(value=msg)


        return exposure_spec

    def read_config_param(self, item, parameter_key):
        """
        Some instrument parameters are dependent on the specifics of detector or instrument setup. This is a generic
        function to pull the correct values out of a dictionary block.

        This is meant to work on dictionaries that are not themselves broken into different filters, gratings, or
        readout patterns; it's for what would otherwise be a flat dictionary of configuration values, like det_pars.

        Values that depend on multiple parameters can be used; the key_format list will contain both parameters and
        the options will contain both relevant values concatenated by "__"

        An example of what a dictionary to be parsed would look like:
        "det_pars": {
            "det_type": "h2rg",
            "dark_current": 0.00636,
            "excessp1": 0.0,
            "excessp2": 0.0,
            "ff_electrons": 62500,
            "fullwell": {
    			"default": 65000.0,
    			"key_format": ["subarray", "readout_pattern"],
    			"full__nrsrapid": 30000.0,
    			"s200a1__nrs": 22000.0
    		},
            "ipc": true,
            "mingroups": 2,
            "pix_size": 18.0,
            "rn": {
    			"default": 16.8,
    			"key_format": ["aperture"],
    			"sw": 17
    		},
            "rn_correlation": true
        }

        item: dict
            Dictionary within which we need to search for the correct detector configuration parameter.
        parameter_key: string
            The parameter we want to search for

        returns: string, int, float, or bool
             The value of parameter_key for the given instrument and detector setup.
        """

        value_lookup = {
            'detector': self.detector_key,
            'aperture': self.instrument['aperture'],
            'subarray': self.detector['subarray'],
            'readout_pattern': self.detector['readout_pattern'],
            'instmode': self.instrument['mode']
        }

        if parameter_key in item:
            if not isinstance(item[parameter_key], dict):
                # If the field exists and is not a dict then just return the value
                parameter_value = item[parameter_key]

            elif isinstance(item[parameter_key], dict) and 'key_format' in item[parameter_key]:
                # If the field exists and it is a dict then we have to check the key_format
                # In this case the format of the parameter will be something like:
                # "rn": {
                #     "default": 16.8,
                #     "key_format": ["aperture"],
                #     "sw": 17
                # },

                # Create the key lookup
                key_lookup = []
                for kf in item[parameter_key]['key_format']:
                    kfelement = kf.split('__')
                    for kfe in kfelement:
                        try:
                            key_lookup.append(value_lookup[kfe])
                        except KeyError as e:
                            raise DataError(
                                "Unknown key format {} value {} listed in {}".format(kfe, value_lookup[kfe],
                                                                                     item[parameter_key]))

                # Now squish the list to a "__" separated key and see if it exists
                # If all required keys were passed in, then create the string lookup.
                # If the string is present in the dict, return it.
                if not any([x is None for x in key_lookup]):
                    key_lookup = '__'.join(key_lookup)

                    if key_lookup in item[parameter_key]:
                        parameter_value = item[parameter_key][key_lookup]

                    # If the key_lookup was not present then just return the default.
                    elif "default" in item[parameter_key]:
                        parameter_value = item[parameter_key]['default']
                    else:
                        raise DataError("There is no default value in {}".format(item[parameter_key]))
            else:
                raise DataError("Improperly formatted parameter {}".format(item))
        else:
            raise DataError("There is no {} key in {}".format(parameter_key, item))

        return parameter_value

    def slitlet_grid(self,slitlet):
        nx = self.slitlet_config[slitlet]['nx']
        ny = self.slitlet_config[slitlet]['ny']
        xstep = self.slitlet_config[slitlet]['xstep']
        ystep = self.slitlet_config[slitlet]['ystep']
        positions = []
        startx = -(nx - 1) / 2.0 * xstep
        stopx = (nx - 1) / 2.0 * xstep
        starty = -(ny - 1) / 2.0 * ystep
        stopy = (ny - 1) / 2.0 * ystep
        startx = int(startx) if startx % 1 == 0 else int(startx - 0.5)
        stopx = int(stopx) if stopx % 1 == 0 else int(stopx - 0.5)
        starty = int(starty) if starty % 1 == 0 else int(starty - 0.5)
        stopy = int(stopy) if stopy % 1 == 0 else int(stopy - 0.5)
        for x in range(startx,stopx+1,xstep):
            for y in range(starty,stopy+1,ystep):
                positions.append([x,y])
        return positions

    def get_aperture_pars(self):
        """
        Collect instrument aperture parameters into a dict and return it

        Returns
        -------
        aperture_dict: dict
            Contains the following parameters describing the instrument aperture:
                disp - size of aperture in dispersion direction (X-axis)
                xdisp - size of aperture perpindicular to dispersion (Y-axis)
                disp_pitch - distance between adjacent apertures in the dispersion direction
                             (None unless a multi-shutter instrument)
                xdisp_pitch - distance between adjacent apertures in the cross-dispersion direction
                              (None unless a multi-shutter instrument)
                pix - pixel scale in arcsec/pixel
                nslice - number of slices the aperture is divided into
        """
        aperture = self.get_aperture()
        if "disp" in self.aperture_config[aperture]:
            disp = self.aperture_config[aperture]["disp"]
            xdisp = self.aperture_config[aperture]["xdisp"]
            nslice = self.aperture_config[aperture]["nslice"]

            """
            For multi-shutter instruments, the shutter aperture size can be smaller than the offset between the
            individual shutters (the pitch). If so, the pitch should be provided in the instrument configuration.
            """
            if "disp_pitch" in self.aperture_config[aperture]:
                try:
                    disp_pitch = self.aperture_config[aperture]['disp_pitch']
                    xdisp_pitch = self.aperture_config[aperture]['xdisp_pitch']
                except KeyError as e:
                    raise DataError("Malformed config file: missing entry for shutter pitch (%s)" % e)
            else:
                """
                If the pitch is not on the config, assume it is the same as the shutter/aperture size.
                """
                disp_pitch = disp
                xdisp_pitch = xdisp

            if "slitlet_shape" in self.instrument:
                if isinstance(self.instrument['slitlet_shape'], str):
                    slitlet_shape = self.slitlet_grid(self.instrument['slitlet_shape'])
                else:
                    slitlet_shape = self.instrument['slitlet_shape']
                """
                The shutter pattern user input is in units of shutter pitch, so convert to internal angular units here.
                """
                multishutter = [(disp_pitch * shutter[0], xdisp_pitch * shutter[1]) for shutter in slitlet_shape]
            else:
                multishutter = None

        else:
            disp = None
            xdisp = None
            disp_pitch = None
            xdisp_pitch = None
            nslice = None
            multishutter = None

        pix = self.aperture_config[aperture]["pix"]

        aperture_dict = {
            'disp': disp,
            'xdisp': xdisp,
            'multishutter': multishutter,
            'pix': pix,
            'nslice': nslice
        }
        return aperture_dict

    def get_wave_range(self):
        """
        Get the wavelength range of the current instrument configuration. The wavelength
        range information is organized in the configuration files on a per aperture basis.

        Returns
        -------
        range_dict: dict
            Contains the instrument wavelength range in microns described by:
                wmin - minimum wavelength
                wmax - maximum wavelength
        """
        disperser = self._get_disperser()
        aperture = self.instrument['aperture']
        filt = self.instrument['filter']

        # if disperser is defined and there are range entries for it, it takes precendence
        # For multiorder projections, we need to scan through a number of dispersion orders, named disperser_n in the
        # ranges section of the configuration file:
        if self.projection_type in "multiorder":
            wmins = []
            wmaxes = []
            for disp in self.range[aperture]:
                # make sure this is one of the disperser's ranges.
                if disperser in disp:
                    if filt is not None and filt in self.range[aperture][disp]: 
                        wmins.append(self.range[aperture][disp][filt]["wmin"])
                        wmaxes.append(self.range[aperture][disp][filt]["wmax"])
                    else:
                        wmins.append(self.range[aperture][disp]["wmin"])
                        wmaxes.append(self.range[aperture][disp]["wmax"])
            # get the min of mins and max of maxes
            wmin = min(wmins)
            wmax = max(wmaxes)
        elif disperser is not None and disperser in self.range[aperture]:
            if filt is not None and filt in self.range[aperture][disperser]:
                wmin = self.range[aperture][disperser][filt]["wmin"]
                wmax = self.range[aperture][disperser][filt]["wmax"]
            else:
                wmin = self.range[aperture][disperser]["wmin"]
                wmax = self.range[aperture][disperser]["wmax"]
        # if filter is defined and has range entries, use it next
        elif filt is not None and filt in self.range[aperture]:
            # Possibility for dependency of the filter on the wavelength range.
            wmin = self.range[aperture][filt]["wmin"]
            wmax = self.range[aperture][filt]["wmax"]
        # otherwise fall back to the generic range for the current aperture
        else:
            wmin = self.range[aperture]["wmin"]
            wmax = self.range[aperture]["wmax"]

        # now check wmin and wmax against wave_pix for sanity, if applicable
        if self.projection_type in ('slitless', 'spec'):
            wave_pix = self.get_wave_pix()
            wmin = max(wmin, wave_pix.min())
            wmax = min(wmax, wave_pix.max())

        range_dict = {"wmin": wmin, "wmax": wmax}
        return range_dict

    @property
    def projection_type(self):
        """
        Determine the appropriate projection type based on the configured instrument mode

        Returns
        -------
        proj_type: str
            The projection type, currently one of 'slitless', 'spec', or 'image'.
        """
        mode = self.mode
        slitless_modes = self.telescope.slitless_modes
        spec_modes = self.telescope.spec_modes
        image_modes = self.telescope.image_modes
        if hasattr(self.telescope, 'multiorder_modes'):
            multiorder_modes = self.telescope.multiorder_modes
        else:
            multiorder_modes = []

        if mode in slitless_modes:
            proj_type = 'slitless'
        if mode in spec_modes:
            proj_type = 'spec'
        if mode in image_modes:
            proj_type = 'image'
        if mode in multiorder_modes:
            proj_type = 'multiorder'
        return proj_type

    def dispersion_axis(self):
        """
        The dispersion axis is either along rows (the X axis) or along columns (Y axis).  By default
        it is along the X axis.  Instrument subclasses will overload this method to deal with cases
        where it is along the Y axis (e.g. NIRISS GR150C).

        Returns
        -------
        disp_axis: str
            Axis along with spectra are dispersed.  Allowed values are 'x' or 'y'.
        """
        disp_axis = "x"
        return disp_axis

    def get_pix_size(self):
        """
        Shortcut to get the size of the pixels in the current observing mode/aperture combination.

        Returns
        -------
        pix_size: float
            The pixel scale in arcsec/pixel
        """
        aperture_dict = self.get_aperture_pars()
        pix_size = aperture_dict['pix']
        return pix_size

    def get_exposure_time(self):
        """
        Shortcut to get the total on-source time for the current detector configuration

        Returns
        -------
        time: float
            Total on-source time in seconds
        """
        time = self.exposure_spec.exposure_time
        return time

    def create_gap_mask(self, wave):
        """
        Use the gap configuration and a wavelength vector, wave, to build a masked array that
        masks out the location of the gap.  Wavelengths between and including both gap endpoints
        will be masked out.

        For now this is a stub and instrument subclasses are expected to overload this to provide
        a gap mask.  Currently only NIRSpec does this.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to construct mask from

        Returns
        -------
        mask: float (default: 1.0)
            1D array masked at the locations within the configured detector gap and 1.0 elsewhere
        """
        return 1.0

    def spectrometer_convolve(self, spec):
        """
        This method convolves a spectrum with a wavelength-dependent spectral
        resolving power R=lambda/delta lambda, where delta lambda is the FWHM of an
        unresolved line at a wavelength lambda. This is accomplished by internally
        changing the sampling of the spectrum, such that delta-lambda is sampled by
        exactly a constant number of samples as a function of wavelength. The convolution
        is then accomplished using the convolution theorem and knowledge that the kernel is
        independent of wavelength.

        Parameters
        ----------
        spec: pandeia.engine.astro_spectrum.AstroSpectrum instance
            The input spectrum to convolve

        Returns
        -------
        spec: pandeia.engine.astro_spectrum.AstroSpectrum instance
            Same dimensions as input spec, but now convolved according to the wavelength-dependent resolving
            power of the instrument, if applicable.  Same as the input spec in the case of no dispersion (i.e. imaging).
        """
        r = self.get_resolving_power(spec.wave)
        if r is None:
            # no dispersion defined so return unmolested spectrum
            return spec

        # find the minimum spacing between wavelengths
        dws = np.abs(spec.wave - np.roll(spec.wave, 1))
        dw_min = np.min(dws)
        dw_arg = np.argmin(dws)

        # check that the minimum spacing is not too small, as this can lead to huge arrays. We will not
        # support sampling better than 2e5 (1e5 is the highest current resolving power of any IR spectrometer).
        if (spec.wave[dw_arg]/dw_min>self.telescope.minimal_spectral_spacing):
            dw_min = spec.wave[dw_arg]/self.telescope.minimal_spectral_spacing
            key = 'spectral_resampling_too_small'
            self.warnings[key] = "Spectrum for source {} has been resampled to a spacing " \
                                 "of {} microns for performance reasons. This is smaller" \
                                 " than the minimum spectral resolution {}.".format(
                self.inst_name,
                spec.wave[dw_arg] / dw_min,
                self.telescope.minimal_spectral_spacing
            )

        fwhm = spec.wave / r  # FWHM of resolution element as a function of wavelength
        fwhm_s = np.min(fwhm / dw_min)  # find min of FWHM expressed in units of minimum spacing (i.e. sampling)

        # but do not allow the sampling FWHM to be less than Nyquist
        fwhm_s = np.max([2., fwhm_s])

        # divide FWHM(wave) by the FWHM of the sampling to get sampling as a function of wavelength
        ds = fwhm / fwhm_s

        # use the min as a starting point; initialize array
        w = np.min(spec.wave)

        # it's much faster (~50%) to append to lists than np.array()'s
        wave_constfwhm = []

        # doing this as a loop is slow, but straightforward.
        while w < np.max(spec.wave):
            # use interpolation to get delta-wavelength from the sampling as a function of wavelength.
            # this method is over 5x faster than the original use of scipy.interpolate.interp1d.
            w += np.interp(w, spec.wave, ds)
            wave_constfwhm.append(w)

        wave_constfwhm.pop()  # remove last point which is an extrapolation
        wave_constfwhm = np.array(wave_constfwhm)

        # interpolate the flux onto the new wavelength set
        flux_constfwhm = spectrum_resample(spec.flux, spec.wave, wave_constfwhm)

        # convolve the flux with a gaussian kernel; first convert the FWHM to sigma
        sigma_s = fwhm_s / 2.3548
        g = Gaussian1DKernel(sigma_s)

        # use boundary='extend' to set values outside the array to nearest array value.
        # this is the best approximation in this case.
        flux_conv = convolve(flux_constfwhm, g, normalize_kernel=True, boundary='extend')
        flux_oldsampling = np.interp(spec.wave, wave_constfwhm, flux_conv)

        spec.flux = flux_oldsampling
        return spec
